from office365.communications.presences.presence import Presence
from office365.directory.extensions.extension import Extension
from office365.directory.licenses.assigned_plan import AssignedPlan
from office365.onedrive.sites.site import Site
from office365.onenote.onenote import Onenote
from office365.outlook.calendar.calendar import Calendar
from office365.outlook.calendar.calendar_group import CalendarGroup
from office365.outlook.calendar.event import Event
from office365.outlook.calendar.meeting_time_suggestions_result import MeetingTimeSuggestionsResult
from office365.outlook.calendar.reminder import Reminder
from office365.directory.licenses.assigned_license import AssignedLicense
from office365.directory.directory_object import DirectoryObject
from office365.directory.directory_object_collection import DirectoryObjectCollection
from office365.directory.licenses.license_details import LicenseDetails
from office365.directory.identities.object_identity import ObjectIdentity
from office365.directory.profile_photo import ProfilePhoto
from office365.entity_collection import EntityCollection
from office365.outlook.contacts.contact import Contact
from office365.outlook.contacts.contact_folder import ContactFolder
from office365.outlook.mail.mailFolder import MailFolder
from office365.onedrive.drives.drive import Drive
from office365.outlook.mail.mailbox_settings import MailboxSettings
from office365.outlook.mail.message import Message
from office365.outlook.outlook_user import OutlookUser
from office365.planner.planner_user import PlannerUser
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.teams.team_collection import TeamCollection


class User(DirectoryObject):
    """Represents an Azure AD user account. Inherits from directoryObject."""

    def assign_license(self, add_licenses, remove_licenses):
        """
        Add or remove licenses on the user.
        :param list[str] remove_licenses: A collection of skuIds that identify the licenses to remove.
        :param list[AssignedLicense] add_licenses: A collection of assignedLicense objects that specify
             the licenses to add.
        """
        params = {
            "addLicenses": ClientValueCollection(AssignedLicense, add_licenses),
            "removeLicenses": ClientValueCollection(str, remove_licenses)
        }
        qry = ServiceOperationQuery(self, "assignLicense", None, params, None, self)
        self.context.add_query(qry)
        return self

    def change_password(self, current_password, new_password):
        """

        :param str current_password:
        :param str new_password:
        """
        qry = ServiceOperationQuery(self, "changePassword", None,
                                    {"currentPassword": current_password, "newPassword": new_password})
        self.context.add_query(qry)
        return self

    def send_mail(self, message, save_to_sent_items=False):
        """Send a new message on the fly

        :type message: office365.mail.message.Message
        :type save_to_sent_items: bool
        """
        payload = {
            "message": message,
            "saveToSentItems": save_to_sent_items
        }
        qry = ServiceOperationQuery(self, "sendmail", None, payload)
        self.context.add_query(qry)
        return self

    def export_personal_data(self, storage_location):
        """
        Submit a data policy operation request from a company administrator or an application to
        export an organizational user's data.

        If successful, this method returns a 202 Accepted response code.
        It does not return anything in the response body. The response contains the following response headers.

        :param str storage_location: This is a shared access signature (SAS) URL to an Azure Storage account,
            to where data should be exported.
        """
        qry = ServiceOperationQuery(self, "exportPersonalData", None, {"storage_location": storage_location})
        self.context.add_query(qry)
        return self

    def find_meeting_times(self):
        """
        Suggest meeting times and locations based on organizer and attendee availability, and time or location
        constraints specified as parameters.

        If findMeetingTimes cannot return any meeting suggestions, the response would indicate a reason in the
        emptySuggestionsReason property. Based on this value, you can better adjust the parameters
        and call findMeetingTimes again.

        The algorithm used to suggest meeting times and locations undergoes fine-tuning from time to time.
        In scenarios like test environments where the input parameters and calendar data remain static, expect
        that the suggested results may differ over time.

        """
        result = ClientResult(self.context, MeetingTimeSuggestionsResult())
        qry = ServiceOperationQuery(self, "findMeetingTimes", None, None, None, result)
        self.context.add_query(qry)
        return result

    def get_calendar_view(self, start_dt, end_dt):
        """Get the occurrences, exceptions, and single instances of events in a calendar view defined by a time range,
           from the user's default calendar, or from some other calendar of the user's.

        :param datetime.datetime end_dt: The end date and time of the time range, represented in ISO 8601 format.
             For example, "2019-11-08T20:00:00-08:00".
        :param datetime.datetime start_dt: The start date and time of the time range, represented in ISO 8601 format.
            For example, "2019-11-08T19:00:00-08:00".

        """
        return_type = EntityCollection(self.context, Event, ResourcePath("calendarView", self.resource_path))
        qry = ServiceOperationQuery(self, "calendarView", None, None, None, return_type)
        self.context.add_query(qry)

        def _construct_request(request):
            """
            :type request: office365.runtime.http.request_options.RequestOptions
            """
            request.method = HttpMethod.Get
            request.url += "?startDateTime={0}&endDateTime={1}".format(start_dt.isoformat(), end_dt.isoformat())

        self.context.before_execute(_construct_request)
        return return_type

    def get_reminder_view(self, start_dt, end_dt):
        """Get the occurrences, exceptions, and single instances of events in a calendar view defined by a time range,
                   from the user's default calendar, or from some other calendar of the user's.

        :param datetime.datetime end_dt: The end date and time of the event for which the reminder is set up.
            The value is represented in ISO 8601 format, for example, "2015-11-08T20:00:00.0000000"..
        :param datetime.datetime start_dt: The start date and time of the event for which the reminder is set up.
            The value is represented in ISO 8601 format, for example, "2015-11-08T19:00:00.0000000".
        """
        result = ClientResult(self.context, ClientValueCollection(Reminder))
        params = {
            "startDateTime": start_dt.isoformat(),
            "endDateTime": end_dt.isoformat(),
        }
        qry = ServiceOperationQuery(self, "reminderView", params, None, None, result)
        self.context.add_query(qry)

        def _construct_request(request):
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_request)
        return result

    def delete_object(self, permanent_delete=False):
        """
        :param permanent_delete: Permanently deletes the user from directory
        :type permanent_delete: bool

        """
        super(User, self).delete_object()
        if permanent_delete:
            deleted_user = self.context.directory.deleted_users[self.id]
            deleted_user.delete_object()
        return self

    def revoke_signin_sessions(self):
        """
        Invalidates all the refresh tokens issued to applications for a user
        (as well as session cookies in a user's browser), by resetting the signInSessionsValidFromDateTime user
        property to the current date-time. Typically, this operation is performed (by the user or an administrator)
        if the user has a lost or stolen device. This operation prevents access to the organization's data through
        applications on the device by requiring the user to sign in again to all applications that they have previously
        consented to, independent of device.
        """
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "revokeSignInSessions", None, None, None, result)
        self.context.add_query(qry)
        return result

    @property
    def account_enabled(self):
        return self.properties.get('accountEnabled', None)

    @property
    def assigned_plans(self):
        """The plans that are assigned to the user."""
        return self.properties.get('assignedPlans', ClientValueCollection(AssignedPlan))

    @property
    def creation_type(self):
        """Indicates whether the user account was created as a regular school or work account (null),
        an external account (Invitation), a local account for an Azure Active Directory B2C tenant (LocalAccount)
        or self-service sign-up using email verification (EmailVerified). Read-only.
        """
        return self.properties.get('creationType', None)

    @property
    def mail(self):
        """The SMTP address for the user, for example, "jeff@contoso.onmicrosoft.com".
           Returned by default. Supports $filter and endsWith.
        """
        return self.properties.get('mail', None)

    @property
    def other_mails(self):
        """A list of additional email addresses for the user;
        for example: ["bob@contoso.com", "Robert@fabrikam.com"]. Supports $filter.
        """
        return self.properties.get('otherMails', ClientValueCollection(str))

    @property
    def identities(self):
        """Represents the identities that can be used to sign in to this user account.
           An identity can be provided by Microsoft (also known as a local account), by organizations,
           or by social identity providers such as Facebook, Google, and Microsoft, and tied to a user account.
           May contain multiple items with the same signInType value.
           Supports $filter.
        """
        return self.properties.get('identities',
                                   ClientValueCollection(ObjectIdentity))

    @property
    def assigned_licenses(self):
        """The licenses that are assigned to the user, including inherited (group-based) licenses. """
        return self.properties.get('assignedLicenses',
                                   ClientValueCollection(AssignedLicense))

    @property
    def followed_sites(self):
        """

        """
        return self.properties.get('followedSites',
                                   EntityCollection(self.context, Site,
                                                    ResourcePath("followedSites", self.resource_path)))

    @property
    def photo(self):
        """
        The user's profile photo. Read-only.
        """
        return self.properties.get('photo',
                                   ProfilePhoto(self.context, ResourcePath("photo", self.resource_path)))

    @property
    def manager(self):
        """
        The user or contact that is this user's manager. Read-only. (HTTP Methods: GET, PUT, DELETE.)
        """
        return self.properties.get('manager',
                                   DirectoryObject(self.context, ResourcePath("manager", self.resource_path)))

    @property
    def preferred_language(self):
        """
        The preferred language for the user. Should follow ISO 639-1 Code; for example en-US.

        :rtype: str or None
        """
        return self.properties.get('preferredLanguage', None)

    @property
    def mailbox_settings(self):
        """
        Get the user's mailboxSettings.

        :rtype: str or None
        """
        return self.properties.get('mailboxSettings', MailboxSettings())

    @property
    def calendar(self):
        """The user's primary calendar. Read-only."""
        return self.properties.get('calendar',
                                   Calendar(self.context, ResourcePath("calendar", self.resource_path)))

    @property
    def calendars(self):
        """The user's calendar groups. Read-only. Nullable."""
        return self.properties.get('calendars',
                                   EntityCollection(self.context, Calendar,
                                                    ResourcePath("calendars", self.resource_path)))

    @property
    def calendar_groups(self):
        """The user's calendar groups. Read-only. Nullable."""
        return self.properties.get('calendarGroups',
                                   EntityCollection(self.context, CalendarGroup,
                                                    ResourcePath("calendarGroups", self.resource_path)))

    @property
    def license_details(self):
        """Retrieve the properties and relationships of a Drive resource."""
        return self.properties.get('licenseDetails',
                                   EntityCollection(self.context, LicenseDetails,
                                                    ResourcePath("licenseDetails", self.resource_path)))

    @property
    def drive(self):
        """Retrieve the properties and relationships of a Drive resource.

        :rtype: Drive
        """
        return self.get_property('drive',
                                 Drive(self.context, ResourcePath("drive", self.resource_path)))

    @property
    def contacts(self):
        """Get a contact collection from the default Contacts folder of the signed-in user (.../me/contacts),
        or from the specified contact folder."""
        return self.properties.get('contacts',
                                   EntityCollection(self.context, Contact,
                                                    ResourcePath("contacts", self.resource_path)))

    @property
    def contact_folders(self):
        """Get the contact folder collection in the default Contacts folder of the signed-in user."""
        return self.properties.get('contactFolders',
                                   EntityCollection(self.context, ContactFolder,
                                                    ResourcePath("contactFolders", self.resource_path)))

    @property
    def events(self):
        """Get an event collection or an event."""
        return self.properties.get('events', EntityCollection(self.context, Event,
                                                              ResourcePath("events", self.resource_path)))

    @property
    def messages(self):
        """Get an event collection or an event."""
        return self.properties.get('messages',
                                   EntityCollection(self.context, Message,
                                                    ResourcePath("messages", self.resource_path)))

    @property
    def joined_teams(self):
        """Get the teams in Microsoft Teams that the user is a direct member of."""
        return self.properties.get('joinedTeams',
                                   TeamCollection(self.context, ResourcePath("joinedTeams", self.resource_path)))

    @property
    def member_of(self):
        """Get groups and directory roles that the user is a direct member of."""
        return self.properties.get('memberOf',
                                   DirectoryObjectCollection(self.context,
                                                             ResourcePath("memberOf", self.resource_path)))

    @property
    def transitive_member_of(self):
        """Get groups, directory roles that the user is a member of. This API request is transitive, and will also
        return all groups the user is a nested member of. """
        return self.properties.get('transitiveMemberOf',
                                   DirectoryObjectCollection(self.context,
                                                             ResourcePath("transitiveMemberOf", self.resource_path)))

    @property
    def mail_folders(self):
        """Get the mail folder collection under the root folder of the signed-in user. """
        return self.properties.get('mailFolders',
                                   EntityCollection(self.context, MailFolder,
                                                    ResourcePath("mailFolders", self.resource_path)))

    @property
    def outlook(self):
        """Represents the Outlook services available to a user."""
        return self.properties.get('outlook',
                                   OutlookUser(self.context, ResourcePath("outlook", self.resource_path)))

    @property
    def onenote(self):
        """Represents the Onenote services available to a user."""
        return self.properties.get('onenote',
                                   Onenote(self.context, ResourcePath("onenote", self.resource_path)))

    @property
    def planner(self):
        """The plannerUser resource provide access to Planner resources for a user."""
        return self.properties.get('planner',
                                   PlannerUser(self.context, ResourcePath("planner", self.resource_path)))

    @property
    def extensions(self):
        """The collection of open extensions defined for the user. Nullable.

        :rtype: EntityCollection
        """
        return self.get_property('extensions',
                                 EntityCollection(self.context, Extension,
                                                  ResourcePath("extensions", self.resource_path)))

    @property
    def direct_reports(self):
        """
        Get a user's direct reports.

        :rtype: EntityCollection
        """
        return self.get_property('directReports',
                                 DirectoryObjectCollection(self.context,
                                                           ResourcePath("directReports", self.resource_path)))

    @property
    def presence(self):
        """Get a user's presence information."""
        return self.properties.get('presence',
                                   Presence(self.context, ResourcePath("presence", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "calendarGroups": self.calendar_groups,
                "contactFolders": self.contact_folders,
                "licenseDetails": self.license_details,
                "memberOf": self.member_of,
                "transitiveMemberOf": self.transitive_member_of,
                "joinedTeams": self.joined_teams,
                "assignedLicenses": self.assigned_licenses,
                "mailFolders": self.mail_folders,
                "mailboxSettings": self.mailbox_settings,
                "directReports": self.direct_reports
            }
            default_value = property_mapping.get(name, None)
        return super(User, self).get_property(name, default_value)

    def set_property(self, name, value, persist_changes=True):
        super(User, self).set_property(name, value, persist_changes)
        # fallback: create a new resource path
        if self._resource_path is None:
            if name == "id" or name == "userPrincipalName":
                self._resource_path = ResourcePath(value, self._parent_collection.resource_path)
        return self



from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.webparts.client_web_part import ClientWebPart


class ClientWebPartCollection(BaseEntityCollection):
    """Web collection"""

    def __init__(self, context, resource_path=None):
        super(ClientWebPartCollection, self).__init__(context, ClientWebPart, resource_path)

    def get_by_id(self, _id):
        """Gets the Client web part with the specified ID."""
        return ClientWebPart(self.context, ResourcePathServiceOperation("getById", [_id], self.resource_path))



from office365.runtime.odata.odata_reader import ODataReader


class ODataV4Reader(ODataReader):
    """OData v4 reader"""
    _options = None

    def __init__(self, metadata_path):
        super(ODataV4Reader, self).__init__(metadata_path, {
            'xmlns': 'http://docs.oasis-open.org/odata/ns/edm',
            'edmx': 'http://docs.oasis-open.org/odata/ns/edmx'
        })




from office365.sharepoint.changes.change import Change


class ChangeList(Change):

    @property
    def base_template(self):
        return self.properties.get("BaseTemplate", None)

    @property
    def list_id(self):
        return self.properties.get("ListId", None)



class FieldType:
    def __init__(self):
        pass

    Invalid = 0
    Integer = 1
    Text = 2
    Note = 3
    DateTime = 4
    Counter = 5
    Choice = 6
    Lookup = 7
    Boolean = 8
    Number = 9
    Currency = 10
    URL = 11
    Computed = 12
    Threading = 13
    Guid = 14
    MultiChoice = 15
    GridChoice = 16
    Calculated = 17
    File = 18
    Attachments = 19
    User = 20
    Recurrence = 21
    CrossProjectLink = 22
    ModStat = 23
    Error = 24
    ContentTypeId = 25
    PageSeparator = 26
    ThreadIndex = 27
    WorkflowStatus = 28
    AllDayEvent = 29
    WorkflowEventType = 30
    Geolocation = 31
    OutcomeChoice = 32
    MaxItems = 33



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.directory.user import User


class DirectorySession(BaseEntity):
    def __init__(self, context):
        super(DirectorySession, self).__init__(context, ResourcePath("SP.Directory.DirectorySession"))

    def me(self):
        user = User(self.context, ResourcePath("me", self.resource_path))
        qry = ServiceOperationQuery(self, "me", None, None, None, user)
        self.context.add_query(qry)
        return user

    @property
    def entity_type_name(self):
        return "SP.Directory.DirectorySession"



from office365.runtime.client_value import ClientValue


class PageLinks(ClientValue):
    """Links for opening a OneNote page."""

    def __init__(self, onenote_client_url=None, onenote_web_url=None):
        """
        :param str onenote_client_url: Opens the page in the OneNote native client if it's installed.
        :param str onenote_web_url: Opens the page in OneNote on the web.
        """
        super(PageLinks, self).__init__()
        self.oneNoteClientUrl = onenote_client_url
        self.oneNoteWebUrl = onenote_web_url



from office365.sharepoint.fields.fieldMultiLookupValue import FieldMultiLookupValue
from office365.sharepoint.fields.field_user_value import FieldUserValue


class FieldMultiUserValue(FieldMultiLookupValue):

    def __init__(self):
        super(FieldMultiUserValue, self).__init__()
        self._item_type = FieldUserValue



from office365.runtime.client_value import ClientValue


class SubwebQuery(ClientValue):
    """Defines a query to specify which child Web sites to return from a Web site."""

    def __init__(self, configuration_filter=-1, web_template_filter=-1):
        super(SubwebQuery, self).__init__()
        """Specifies the site definition."""
        self.WebTemplateFilter = web_template_filter
        """ A 16-bit integer that specifies the identifier of a configuration"""
        self.ConfigurationFilter = configuration_filter



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.audit.audit import Audit
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.changes.change_collection import ChangeCollection
from office365.sharepoint.eventreceivers.event_receiver_definition import EventReceiverDefinitionCollection
from office365.sharepoint.features.feature_collection import FeatureCollection
from office365.sharepoint.lists.list import List
from office365.sharepoint.portal.site_icon_manager import SiteIconManager
from office365.sharepoint.principal.user import User
from office365.sharepoint.recyclebin.recycleBinItemCollection import RecycleBinItemCollection
from office365.sharepoint.sites.site_health_summary import SiteHealthSummary
from office365.sharepoint.sites.sph_site import SPHSite
from office365.sharepoint.sites.usage_info import UsageInfo
from office365.sharepoint.usercustomactions.user_custom_action_collection import UserCustomActionCollection
from office365.sharepoint.webs.web import Web
from office365.sharepoint.webs.web_template_collection import WebTemplateCollection
from office365.sharepoint.types.resource_path import ResourcePath as SPResPath


class Site(BaseEntity):
    """Represents a collection of sites in a Web application, including a top-level website and all its sub sites."""

    def __init__(self, context):
        super(Site, self).__init__(context, ResourcePath("Site", None))

    @staticmethod
    def from_url(url):
        """Construct and return a site instance

        :type url: str
        """
        from office365.sharepoint.client_context import ClientContext
        client = ClientContext(url)
        return client.site

    def get_site_logo(self):
        """
        Downloads a site logo
        """
        return_type = ClientResult(self.context)

        def _site_loaded():
            site_manager = SiteIconManager(self.context)
            site_manager.get_site_logo(self.url, return_type=return_type)

        self.ensure_property("Url", _site_loaded)
        return return_type

    def set_site_logo(self, relative_logo_url):
        """Uploads a site logo

        :param str relative_logo_url:
        """
        site_manager = SiteIconManager(self.context)
        site_manager.set_site_logo(relative_logo_url=relative_logo_url)
        return self

    def is_valid_home_site(self):
        return_type = ClientResult(self.context)

        def _site_loaded():
            SPHSite.is_valid_home_site(self.context, self.url, return_type)

        self.ensure_property("Url", _site_loaded)
        return return_type

    def set_as_home_site(self):
        result = ClientResult(self.context)

        def _site_loaded():
            self.result = SPHSite.set_as_home_site(self.context, self.url, result)

        self.ensure_property("Url", _site_loaded)
        return result

    def get_changes(self, query):
        """Returns the collection of all changes from the change log that have occurred within the scope of the site,
        based on the specified query.

        :param office365.sharepoint.changes.change_query.ChangeQuery query: Specifies which changes to return
        """
        changes = ChangeCollection(self.context)
        qry = ServiceOperationQuery(self, "getChanges", None, query, "query", changes)
        self.context.add_query(qry)
        return changes

    def get_recycle_bin_items(self, row_limit=100, is_ascending=True):
        """
        Returns a collection of recycle bin items based on the specified query.

        :param int row_limit:
        :param bool is_ascending:
        """
        result = RecycleBinItemCollection(self.context)
        payload = {
            "rowLimit": row_limit,
            "isAscending": is_ascending
        }
        qry = ServiceOperationQuery(self, "GetRecycleBinItems", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def get_web_templates(self, lcid=1033, override_compat_level=0):
        """
        Returns the collection of site definitions that are available for creating
            Web sites within the site collection.<99>

        :param int lcid: A 32-bit unsigned integer that specifies the language of the site definitions that are
            returned from the site collection.
        :param int override_compat_level: Specifies the compatibility level of the site (2)
            to return from the site collection. If this value is 0, the compatibility level of the site (2) is used.
        :return:
        """
        params = {
            "LCID": lcid,
            "overrideCompatLevel": override_compat_level
        }
        return_type = WebTemplateCollection(self.context,
                                            ResourcePathServiceOperation("GetWebTemplates", params, self.resource_path))

        qry = ServiceOperationQuery(self, "GetWebTemplates", params, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    def join_hub_site(self, hub_site_id, approval_token, approval_correlation_id):
        """

        """
        params = {
            "hubSiteId": hub_site_id,
            "approvalToken": approval_token,
            "approvalCorrelationId": approval_correlation_id
        }
        return_type = WebTemplateCollection(self.context,
                                            ResourcePathServiceOperation("GetWebTemplates", params, self.resource_path))

        qry = ServiceOperationQuery(self, "JoinHubSite", params, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    @staticmethod
    def get_url_by_id(context, site_id, stop_redirect=False):
        """Gets Site Url By Id

        :type context: office365.sharepoint.client_context.ClientContext
        :type site_id: str
        :type stop_redirect: bool
        """
        result = ClientResult(context)
        payload = {
            "id": site_id,
            "stopRedirect": stop_redirect
        }
        qry = ServiceOperationQuery(context.site, "GetUrlById", None, payload, None, result)
        qry.static = True
        context.add_query(qry)
        return result

    @staticmethod
    def get_url_by_id_for_web(context, site_id, stop_redirect, web_id):
        """Gets Site Url By Id

        :type context: office365.sharepoint.client_context.ClientContext
        :type site_id: str
        :type stop_redirect: bool
        :type web_id: str
        """
        result = ClientResult(context)
        payload = {
            "id": site_id,
            "stopRedirect": stop_redirect,
            "webId": web_id
        }
        qry = ServiceOperationQuery(context.site, "GetUrlById", None, payload, None, result)
        qry.static = True
        context.add_query(qry)
        return result

    @staticmethod
    def exists(context, url):
        """Determine whether site exists
        :type context: office365.sharepoint.client_context.ClientContext
        :type url: str
        """
        result = ClientResult(context)
        payload = {
            "url": url
        }
        qry = ServiceOperationQuery(context.site, "Exists", None, payload, None, result)
        qry.static = True
        context.add_query(qry)
        return result

    def get_catalog(self, type_catalog):
        """
        Specifies the list template gallery, site template gallery, Web Part gallery, master page gallery,
        or other galleries from the site collection, including custom galleries that are defined by users.

        :type type_catalog: int
        """
        return List(self.context, ResourcePathServiceOperation("getCatalog", [type_catalog], self.resource_path))

    def open_web(self, str_url):
        """Returns the specified Web site from the site collection.

        :param str str_url: A string that contains either the server-relative or site-relative URL of the
        Web site or of an object within the Web site. A server-relative URL begins with a forward slash ("/"),
        while a site-relative URL does not begin with a forward slash.
        """
        return_type = Web(self.context)
        qry = ServiceOperationQuery(self, "OpenWeb", {"strUrl": str_url}, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    def open_web_by_id(self, web_id):
        """Returns the specified Web site from the site collection.

        :param str web_id: An identifier of the Web site
        """
        return_type = Web(self.context)
        qry = ServiceOperationQuery(self, "OpenWebById", {"gWebId": web_id}, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    def register_hub_site(self, create_info):
        """Registers an existing site as a hub site.

        :type create_info: HubSiteCreationInformation
        """
        qry = ServiceOperationQuery(self, "RegisterHubSite", None, create_info, "creationInformation", None)
        self.context.add_query(qry)
        return self

    def run_health_check(self, rule_id, repair, run_always):
        """
        :param str rule_id:
        :param bool repair:
        :param bool run_always:
        """
        payload = {
            "ruleId": rule_id,
            "bRepair": repair,
            "bRunAlways": run_always
        }
        return_type = SiteHealthSummary(self.context)
        qry = ServiceOperationQuery(self, "RunHealthCheck", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def unregister_hub_site(self):
        """
        Disables the hub site feature on a site.
        """
        qry = ServiceOperationQuery(self, "UnRegisterHubSite", None, None, None, None)
        self.context.add_query(qry)
        return self

    @property
    def audit(self):
        """
        Enables auditing of how site collection is accessed, changed, and used.
        """
        return self.properties.get("Audit", Audit(self.context, ResourcePath("Audit", self.resource_path)))

    @property
    def root_web(self):
        """Get root web"""
        return self.properties.get('RootWeb', Web(self.context, ResourcePath("RootWeb", self.resource_path)))

    @property
    def owner(self):
        """Gets or sets the owner of the site collection. (Read-only in sandboxed solutions.)"""
        return self.properties.get('owner', User(self.context, ResourcePath("owner", self.resource_path)))

    @property
    def read_only(self):
        """
        Gets a Boolean value that specifies whether the site collection is read-only,
        locked, and unavailable for write access.

        :rtype: bool
        """
        return self.properties.get("ReadOnly", None)

    @property
    def required_designer_version(self):
        """
        :rtype: str
        """
        return self.properties.get("RequiredDesignerVersion", None)

    @property
    def url(self):
        """
        Site url

        :rtype: str
        """
        return self.properties.get('Url', None)

    def server_relative_url(self):
        """
        Gets the server-relative URL of the root Web site in the site collection.

        :rtype: str
        """
        return self.properties.get("ServerRelativeUrl", None)

    @property
    def id(self):
        """
        :rtype: str
        """
        return self.properties.get("Id", None)

    @property
    def is_hub_site(self):
        """
        :rtype: bool
        """
        return self.properties.get("IsHubSite", None)

    @property
    def server_relative_path(self):
        """Gets the server-relative Path of the Site.

        :rtype: SPResPath or None
        """
        return self.properties.get("ServerRelativePath", SPResPath())

    @property
    def secondary_contact(self):
        """Gets or sets the secondary contact that is used for the site collection."""
        return self.properties.get('SecondaryContact', User(self.context,
                                                            ResourcePath("SecondaryContact", self.resource_path)))

    @property
    def recycle_bin(self):
        """Get recycle bin"""
        return self.properties.get('RecycleBin',
                                   RecycleBinItemCollection(self.context,
                                                            ResourcePath("RecycleBin", self.resource_path)))

    @property
    def features(self):
        """Get features"""
        return self.properties.get('Features',
                                   FeatureCollection(self.context,
                                                     ResourcePath("Features", self.resource_path), self))

    @property
    def event_receivers(self):
        """
        Provides event receivers for events that occur at the scope of the site collection.
        """
        return self.properties.get('EventReceivers',
                                   EventReceiverDefinitionCollection(self.context,
                                                                     ResourcePath("eventReceivers", self.resource_path),
                                                                     self))

    @property
    def usage_info(self):
        """Provides fields used to access information regarding site collection usage."""
        return self.properties.get("UsageInfo", UsageInfo())

    @property
    def user_custom_actions(self):
        """Gets the User Custom Actions that are associated with the site."""
        return self.properties.get('UserCustomActions',
                                   UserCustomActionCollection(self.context,
                                                              ResourcePath("UserCustomActions", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "EventReceivers": self.event_receivers,
                "RecycleBin": self.recycle_bin,
                "RootWeb": self.root_web,
                "SecondaryContact": self.secondary_contact,
                "UsageInfo": self.usage_info,
                "UserCustomActions": self.user_custom_actions
            }
            default_value = property_mapping.get(name, None)
        return super(Site, self).get_property(name, default_value)



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.sites.language_collection import LanguageCollection


class ServerSettings(BaseEntity):
    """Provides methods for obtaining server properties."""

    def __init__(self, context):
        super(ServerSettings, self).__init__(context, ResourcePath("SP.ServerSettings"))

    @staticmethod
    def is_sharepoint_online(context):
        """
        :type context: office365.sharepoint.client_context.ClientContext
        """
        binding_type = ServerSettings(context)
        return_type = ClientResult(context)
        qry = ServiceOperationQuery(binding_type, "IsSharePointOnline", None, None, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def get_blocked_file_extensions(context):
        """
        :type context: office365.sharepoint.client_context.ClientContext
        """
        binding_type = ServerSettings(context)
        return_type = ClientResult(context, ClientValueCollection(str))
        qry = ServiceOperationQuery(binding_type, "GetBlockedFileExtensions", None, None, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def get_global_installed_languages(context, compatibility_level):
        """
        Gets a list of installed languages that are compatible with a given version of SharePoint.

        :type context: office365.sharepoint.client_context.ClientContext
        :param int compatibility_level: The value of the major SharePoint version to query for installed languages.
        """
        binding_type = ServerSettings(context)
        return_type = LanguageCollection(context)
        qry = ServiceOperationQuery(binding_type, "GetGlobalInstalledLanguages", [compatibility_level], None, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.principal.principal import Principal
from office365.sharepoint.utilities.principal_info import PrincipalInfo


class Group(Principal):
    """Represents a collection of users in a SharePoint site. A group is a type of SP.Principal."""

    def expand_to_principals(self, max_count=10):
        return_type = ClientResult(self.context, ClientValueCollection(PrincipalInfo))

        def _group_loaded():
            from office365.sharepoint.utilities.utility import Utility
            Utility.expand_groups_to_principals(self.context, [self.login_name], max_count, return_type)
        self.ensure_property("LoginName", _group_loaded)
        return return_type

    @property
    def owner_title(self):
        """Specifies the name of the owner of the group.
        :rtype: str or None
        """
        return self.properties.get('OwnerTitle', None)

    @property
    def can_current_user_manage_group(self):
        """Gets a Boolean value that indicates whether the current user can manage the group.

        :rtype: bool or None
        """
        return self.properties.get('CanCurrentUserManageGroup', None)

    @property
    def owner(self):
        """Specifies the owner of the group."""
        return self.properties.get('Owner', Principal(self.context, ResourcePath("Owner", self.resource_path)))

    @property
    def users(self):
        """Gets a collection of user objects that represents all of the users in the group."""
        from office365.sharepoint.principal.user_collection import UserCollection
        return self.properties.get('Users', UserCollection(self.context, ResourcePath("Users", self.resource_path)))



from office365.entity_collection import EntityCollection
from office365.outlook.calendar.attendee import Attendee
from office365.directory.extensions.extension import Extension
from office365.outlook.calendar.dateTimeTimeZone import DateTimeTimeZone
from office365.outlook.calendar.email_address import EmailAddress
from office365.outlook.mail.attachment_collection import AttachmentCollection
from office365.outlook.mail.item import Item
from office365.outlook.mail.itemBody import ItemBody
from office365.outlook.mail.location import Location
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.resource_path import ResourcePath


class Event(Item):
    """An event in a user calendar, or the default calendar of a Microsoft 365 group."""

    @property
    def start(self):
        """
        The date, time, and time zone that the event starts. By default, the start time is in UTC.

        :rtype: DateTimeTimeZone
        """
        return self.get_property("start", DateTimeTimeZone())

    @start.setter
    def start(self, value):
        """
        Sets the date, time, and time zone that the event starts. By default, the start time is in UTC.
        """
        self.set_property("start", DateTimeTimeZone.parse(value))

    @property
    def end(self):
        """
        The date, time, and time zone that the event starts. By default, the start time is in UTC.

        :rtype: DateTimeTimeZone
        """
        return self.get_property("end", DateTimeTimeZone())

    @end.setter
    def end(self, value):
        """
        Sets the date, time, and time zone that the event starts. By default, the start time is in UTC.
        """
        self.set_property("end", DateTimeTimeZone.parse(value))

    @property
    def body(self):
        """
        The body of the message associated with the event. It can be in HTML or text format.
        """
        return self.get_property("body", ItemBody())

    @body.setter
    def body(self, value):
        """
        Sets The body of the message associated with the event. It can be in HTML or text format.
        """
        self.set_property("body", ItemBody(value, "HTML"))

    @property
    def body_preview(self):
        """
        The preview of the message associated with the event. It is in text format.
        :rtype: str or None
        """
        return self.properties.get("bodyPreview", None)

    @property
    def subject(self):
        """
        The text of the event's subject line.
        :rtype: str or None
        """
        return self.properties.get("subject", None)

    @subject.setter
    def subject(self, value):
        """
        Sets The text of the event's subject line.
        :type: str or None
        """
        self.set_property("subject", value)

    @property
    def location(self):
        """
        The location of the event.
        """
        return self.properties.get("location", Location())

    @property
    def web_link(self):
        """
        The URL to open the event in Outlook on the web.

        Outlook on the web opens the event in the browser if you are signed in to your mailbox. Otherwise, Outlook
        on the web prompts you to sign in.

        This URL cannot be accessed from within an iFrame.

        :rtype: str or None
        """
        return self.properties.get("webLink", None)

    @property
    def calendar(self):
        """The calendar that contains the event. Navigation property. Read-only."""
        from office365.outlook.calendar.calendar import Calendar
        return self.properties.get('calendar',
                                   Calendar(self.context, ResourcePath("calendar", self.resource_path)))

    @property
    def attendees(self):
        """The collection of attendees for the event."""
        return self.properties.get('attendees', ClientValueCollection(Attendee))

    @attendees.setter
    def attendees(self, value):
        """Sets the collection of attendees for the event.

        :type value: list[str]
        """
        self.set_property('attendees',
                          ClientValueCollection(Attendee,
                                                [Attendee(EmailAddress(v), attendee_type="required") for v in value]))

    @property
    def attachments(self):
        """The collection of fileAttachment and itemAttachment attachments for the event. """
        return self.properties.get('attachments',
                                   AttachmentCollection(self.context, ResourcePath("attachments", self.resource_path)))

    @property
    def extensions(self):
        """The collection of open extensions defined for the event. Nullable."""
        return self.properties.get('extensions',
                                   EntityCollection(self.context, Extension,
                                                    ResourcePath("extensions", self.resource_path)))

    @property
    def instances(self):
        """The collection of open extensions defined for the event. Nullable."""
        return self.properties.get('instances',
                                   EntityCollection(self.context, Event, ResourcePath("instances", self.resource_path)))



import os

from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.folders.folder import Folder


class FolderCollection(BaseEntityCollection):
    """Represents a collection of Folder resources."""
    def __init__(self, context, resource_path=None, parent=None):
        super(FolderCollection, self).__init__(context, Folder, resource_path, parent)

    def get(self):
        """
        :rtype: FolderCollection
        """
        return super(FolderCollection, self).get()

    def add_using_path(self, decoded_url, overwrite):
        """
        :type decoded_url: str
        :type overwrite:  bool
        """
        parameters = {
            "DecodedUrl": decoded_url,
            "Overwrite": overwrite
        }
        target_folder = Folder(self.context)
        qry = ServiceOperationQuery(self, "AddUsingPath", parameters, None, None, target_folder)
        self.context.add_query(qry)
        return target_folder

    def get_by_path(self, decoded_url):
        """

        :type decoded_url: str
        """
        from office365.sharepoint.types.resource_path import ResourcePath as SPResPath
        target_folder = Folder(self.context)
        qry = ServiceOperationQuery(self, "GetByPath", SPResPath(decoded_url), None, "parameters", target_folder)
        self.context.add_query(qry)
        return target_folder

    def ensure_folder_path(self, path):
        """
        Ensures a nested folder hierarchy exist

        :param str path: relative server URL (path) to a folder
        """

        url_component = os.path.normpath(path).split(os.path.sep)
        url_component = [part for part in url_component if part]
        if not url_component:
            raise NotADirectoryError("Wrong relative URL provided")
        child_folder = self
        for url_part in url_component:
            child_folder = child_folder.add(url_part)
        return child_folder

    def add(self, server_relative_url):
        """Adds the folder that is located at the specified URL to the collection.

        :type server_relative_url: str
        """
        folder = Folder(self.context)
        self.add_child(folder)
        folder.set_property("ServerRelativeUrl", server_relative_url)
        qry = CreateEntityQuery(self, folder, folder)
        self.context.add_query(qry)
        return folder

    def get_by_url(self, url):
        """Retrieve Folder resource by url
        :type url: str
        """
        return Folder(self.context, ResourcePathServiceOperation("GetByUrl", [url], self.resource_path))



from office365.entity import Entity


class GroupSetting(Entity):
    """Group settings control behaviors such as blocked word lists for group display names or whether guest users are
    allowed to be group owners. """



from office365.outlook.outlook_entity import OutlookEntity


class Item(OutlookEntity):
    """"""

    @property
    def change_key(self):
        """Identifies the version of the mail object. Every time the event is changed, ChangeKey changes as well.
           This allows Exchange to apply changes to the correct version of the object."""
        return self.properties.get('ChangeKey', None)



from office365.runtime.client_value import ClientValue


class FileCreationInformation(ClientValue):
    """Represents properties that can be set when creating a file by using the FileCollection.Add method."""

    def __init__(self, url=None, overwrite=False, content=None):
        """

        :type url: str
        """
        super(FileCreationInformation, self).__init__()
        self._url = url
        self._overwrite = overwrite
        self._content = content

    def to_json(self, json_format=None):
        return {
            "overwrite": self.overwrite,
            "url": self.url
        }

    @property
    def content(self):
        """Gets the binary content of the file."""
        return self._content

    @content.setter
    def content(self, value):
        """Sets the binary content of the file."""
        self._content = value

    @property
    def overwrite(self):
        """Indicates whether to overwrite an existing file with the same name and in the same location
        as the one being added."""
        return self._overwrite

    @property
    def url(self):
        """The URL of the file."""
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @overwrite.setter
    def overwrite(self, value):
        self._overwrite = value



from abc import abstractmethod

import requests
from requests import HTTPError

from office365.runtime.client_request_exception import ClientRequestException
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.types.EventHandler import EventHandler


class ClientRequest(object):

    def __init__(self, context):
        """
        Abstract request client

        :type context: office365.runtime.client_runtime_context.ClientRuntimeContext
        """
        self._context = context
        self._queries = []
        self._current_query = None
        self.beforeExecute = EventHandler()
        self.afterExecute = EventHandler()

    @property
    def context(self):
        return self._context

    @property
    def queries(self):
        """
        :rtype: list[ClientQuery]
        """
        return self._queries

    @property
    def current_query(self):
        """
        :rtype: office365.runtime.queries.client_query.ClientQuery
        """
        return self._current_query

    def add_query(self, query, execute_first=False, set_as_current=True, reset_queue=False):
        """
        :type query: office365.runtime.queries.client_query.ClientQuery
        :type execute_first: bool
        :type set_as_current: bool
        :type reset_queue: bool
        """
        if reset_queue:
            self._queries = []
        if set_as_current:
            self._current_query = query
        if execute_first:
            self._queries.insert(0, query)
        else:
            self._queries.append(query)

    def clear(self):
        self._current_query = None
        self._queries = []

    def build_single_request(self, query):
        """

        :type: office365.runtime.queries.client_query.ClientQuery
        """
        self._current_query = query
        return self.build_request()

    @abstractmethod
    def build_request(self):
        """
        :rtype: office365.runtime.http.request_options.RequestOptions
        """
        pass

    @abstractmethod
    def process_response(self, response):
        """
        :type response: requests.Response
        """
        pass

    def execute_query(self,**kwargs):
        """
        Submit a pending request to the server
        """
        for _ in self.next_query():
            try:
                request = self.build_request()
                self.beforeExecute.notify(request)
                response = self.execute_request_direct(request,**kwargs)
                response.raise_for_status()
                self.process_response(response)
                self.afterExecute.notify(response)
            except HTTPError as e:
                raise ClientRequestException(*e.args, response=e.response)

    def execute_request_direct(self, request_options,timeout_=None):
        """Execute client request

        :type request_options: office365.runtime.http.request_options.RequestOptions
        """
        self.context.authenticate_request(request_options)
        if request_options.method == HttpMethod.Post:
            if request_options.is_bytes or request_options.is_file:
                response = requests.post(url=request_options.url,
                                         headers=request_options.headers,
                                         data=request_options.data,
                                         auth=request_options.auth,
                                         verify=request_options.verify,
                                         timeout=timeout_)
            else:
                response = requests.post(url=request_options.url,
                                         headers=request_options.headers,
                                         json=request_options.data,
                                         auth=request_options.auth,
                                         verify=request_options.verify,
                                         timeout=timeout_)
        elif request_options.method == HttpMethod.Patch:
            response = requests.patch(url=request_options.url,
                                      headers=request_options.headers,
                                      json=request_options.data,
                                      auth=request_options.auth,
                                      verify=request_options.verify,
                                      timeout=timeout_)
        elif request_options.method == HttpMethod.Delete:
            response = requests.delete(url=request_options.url,
                                       headers=request_options.headers,
                                       auth=request_options.auth,
                                       verify=request_options.verify,
                                       timeout=timeout_)
        elif request_options.method == HttpMethod.Put:
            response = requests.put(url=request_options.url,
                                    data=request_options.data,
                                    headers=request_options.headers,
                                    auth=request_options.auth,
                                    verify=request_options.verify,
                                    timeout=timeout_)
        else:
            response = requests.get(url=request_options.url,
                                    headers=request_options.headers,
                                    auth=request_options.auth,
                                    verify=request_options.verify,
                                    stream=request_options.stream,
                                    proxies=request_options.proxies,
                                    timeout=timeout_)
        return response

    def next_query(self):
        while len(self._queries) > 0:
            qry = self._queries.pop(0)
            self._current_query = qry
            yield qry



from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.resource_path import ResourcePath
from office365.teams.messages.chat_message_attachment import ChatMessageAttachment


class ChatMessage(Entity):
    """
    Represents an individual chat message within a channel or chat.
    The message can be a root message or part of a thread that is defined by the replyToId property in the message.
    """

    @property
    def attachments(self):
        """The collection of replies."""
        return self.properties.get("attachments", ClientValueCollection(ChatMessageAttachment()))

    @property
    def replies(self):
        """
        The collection of replies.
        """
        return self.properties.get("replies",
                                   EntityCollection(self.context, ChatMessage,
                                                    ResourcePath("replies", self.resource_path)))

    @property
    def web_url(self):
        """
        :rtype: str
        """
        return self.properties.get("webUrl", None)

    @property
    def importance(self):
        """
        The importance of the chat message. The possible values are: normal, high, urgent.
        :rtype: str
        """
        return self.properties.get("importance", None)



from xml.etree import ElementTree as ET

from office365.runtime.odata.odata_model import ODataModel
from office365.runtime.odata.odata_property import ODataProperty
from office365.runtime.odata.odata_type import ODataType


class ODataReader(object):
    """OData reader"""

    def __init__(self, metadata_path, xml_namespaces):
        """
        :type metadata_path: string
        :type xml_namespaces: dict
        """
        self._metadata_path = metadata_path
        self._xml_namespaces = xml_namespaces

    def format_file(self):
        import xml.dom.minidom
        with open(self._metadata_path, 'r') as in_file:
            metadata_content = in_file.read()

        formatted_metadata_content = xml.dom.minidom.parseString(metadata_content).toprettyxml()

        with open(self._metadata_path, "w", encoding="utf8") as out_file:
            out_file.write(formatted_metadata_content)

    def process_schema_node(self, model):
        """
        :type model: ODataModel
        """
        root = ET.parse(self._metadata_path).getroot()
        schema_node = root.find('edmx:DataServices/xmlns:Schema', self._xml_namespaces)
        for type_node in schema_node.findall('xmlns:ComplexType', self._xml_namespaces):
            type_schema = self.process_type_node(type_node, schema_node)
            model.add_type(type_schema)

    def process_type_node(self, type_node, schema_node):
        """
        :type type_node: xml.etree.ElementTree.Element
        :type schema_node: xml.etree.ElementTree.Element
        """
        type_schema = ODataType()
        type_schema.namespace = schema_node.attrib['Namespace']
        type_schema.name = type_node.get('Name')
        type_schema.baseType = 'ComplexType'

        for prop_node in type_node.findall('xmlns:Property', self._xml_namespaces):
            prop_schema = self.process_property_node(prop_node)
            type_schema.add_property(prop_schema)

        return type_schema

    def process_method_node(self):
        pass

    def process_property_node(self, node):
        """
        :type node:  xml.etree.ElementTree.Element
        """
        prop_schema = ODataProperty()
        prop_schema.name = node.get('Name')
        return prop_schema

    def generate_model(self):
        model = ODataModel()
        self.process_schema_node(model)
        return model



import os
from abc import abstractmethod

from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.client_query import ClientQuery


class UploadSessionQuery(ClientQuery):
    """Create an upload session to allow your app to upload files up to the maximum file size. An upload session
    allows your app to upload ranges of the file in sequential API requests, which allows the transfer to be resumed
    if a connection is dropped while the upload is in progress. """

    def __init__(self, session_item, source_path, chunk_size=1024, chunk_uploaded=None):
        """

        :type session_item: office365.entity.Entity
        :type source_path: str
        :type chunk_size: int
        """
        super(UploadSessionQuery, self).__init__(session_item.context, session_item)
        self._session_item = session_item
        self._file_handle = None
        self._chunk_size = chunk_size
        self._range_start = 0
        self._range_end = 0
        self._chunk_uploaded = chunk_uploaded
        self._source_path = source_path
        self._session_result = None
        self._session_result = self.create_upload_session()
        self.context.after_execute(self._create_next_range_query)

    @abstractmethod
    def create_upload_session(self):
        pass

    def _create_next_range_query(self, resp):
        """
        :type resp: requests.Response
        """
        if self._has_pending_read():
            qry = ClientQuery(self.context, self.return_type)
            self.context.before_execute(self._construct_range_request)
            self.context.after_execute(self._create_next_range_query)
            self.context.after_execute(self._notify_after_uploaded)
            self.context.add_query(qry, True)

    def _construct_range_request(self, request):
        range_data = self._read_next_chunk()
        request.url = self._session_result.value.uploadUrl
        request.method = HttpMethod.Put
        request.set_header('Content-Length', str(len(range_data)))
        request.set_header('Content-Range',
                           'bytes {0}-{1}/{2}'.format(self._range_start, self._range_end - 1, self.file_size))
        request.set_header('Accept', '*/*')
        request.data = range_data

    def _notify_after_uploaded(self, response):
        response.raise_for_status()
        if callable(self._chunk_uploaded):
            self._chunk_uploaded(self._range_end)

    def _read_next_chunk(self):
        self._range_start = self._file_handle.tell()
        range_data = self._file_handle.read(self._chunk_size)
        self._range_end = self._file_handle.tell()
        return range_data

    def _has_pending_read(self):
        if self._range_end >= self.file_size:
            if self._file_handle is not None:
                self._file_handle.close()
            return False
        if self._file_handle is None:
            self._file_handle = open(self._source_path, 'rb')
        return True

    @property
    def file_size(self):
        return os.stat(self._source_path).st_size

    @property
    def file_name(self):
        return os.path.basename(self._source_path)



from office365.runtime.client_value import ClientValue


class Audio(ClientValue):
    """
    The Audio resource groups audio-related properties on an item into a single structure.

    If a DriveItem has a non-null audio facet, the item represents an audio file.
    The properties of the Audio resource are populated by extracting metadata from the file.
    """
    pass



from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection


class SecondaryAdministratorsFieldsData(ClientValue):

    def __init__(self, site_id, emails=None, names=None):
        """
        :type emails: List[str] or None
        :type names: List[str] or None
        :type site_id: str or None
        """
        super(SecondaryAdministratorsFieldsData, self).__init__()
        self.secondaryAdministratorEmails = ClientValueCollection(str, emails)
        self.secondaryAdministratorLoginNames = ClientValueCollection(str, names)
        self.siteId = site_id

    @property
    def entity_type_name(self):
        return "Microsoft.Online.SharePoint.TenantAdministration.SecondaryAdministratorsFieldsData"



from office365.entity_collection import EntityCollection
from office365.onenote.entity_hierarchy_model import OnenoteEntityHierarchyModel
from office365.onenote.sections.section import OnenoteSection
from office365.runtime.resource_path import ResourcePath


class Notebook(OnenoteEntityHierarchyModel):
    """A OneNote notebook."""

    @property
    def sections(self):
        """Retrieve a list of onenoteSection objects from the specified notebook.

        :rtype: EntityCollection
        """
        return self.get_property('sections', EntityCollection(self.context, OnenoteSection,
                                                              ResourcePath("sections", self.resource_path)))

    @property
    def section_groups(self):
        """Retrieve a list of onenoteSection objects from the specified notebook.

        :rtype: EntityCollection
        """
        from office365.onenote.sectiongroups.section_group import SectionGroup
        return self.get_property('sectionGroups',
                                 EntityCollection(self.context, SectionGroup,
                                                  ResourcePath("sectionGroups", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "sectionGroups": self.section_groups
            }
            default_value = property_mapping.get(name, None)
        return super(Notebook, self).get_property(name, default_value)



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.fields.related_field import RelatedField


class RelatedFieldCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(RelatedFieldCollection, self).__init__(context, RelatedField, resource_path)



from office365.runtime.client_value import ClientValue


class PhysicalAddress(ClientValue):
    """The physical address of a contact."""

    def __init__(self, city=None, country_or_region=None, postal_code=None):
        """
        :param str city: strThe city.
        :param str country_or_region: The country or region. It's a free-format string value, for example, "United States".
        :param str postal_code: The postal code.
        """
        super(PhysicalAddress, self).__init__()
        self.city = city
        self.countryOrRegion = country_or_region
        self.postalCode = postal_code



from office365.runtime.client_value import ClientValue


class SearchRequest(ClientValue):

    def __init__(self, query_text, **kwargs):
        super(SearchRequest, self).__init__()
        self.Querytext = query_text
        self.ClientType = None
        self.CollapseSpecification = None
        self.Culture = None
        self.__dict__.update(**kwargs)

    @property
    def entity_type_name(self):
        return "Microsoft.Office.Server.Search.REST.SearchRequest"



from office365.runtime.client_object_collection import ClientObjectCollection
from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.principal.user import User
from office365.sharepoint.tenant.management.users_results import GetExternalUsersResults, RemoveExternalUsersResults, \
    SPOUserSessionRevocationResult
from office365.sharepoint.tenant.administration.theme_properties import ThemeProperties


class Office365Tenant(BaseEntity):
    """Represents a SharePoint Online tenant."""

    @property
    def allow_editing(self):
        return self.properties.get("AllowEditing", None)

    def __init__(self, context):
        super(Office365Tenant, self).__init__(
            context,
            ResourcePath("Microsoft.Online.SharePoint.TenantManagement.Office365Tenant")
        )

    def add_tenant_cdn_origin(self, cdn_type, origin_url):
        """
        Configures a new origin to public or private CDN, on either Tenant level or on a single Site level.
        Effectively, a tenant admin points out to a document library, or a folder in the document library
        and requests that content in that library should be retrievable by using a CDN.

        You must have the SharePoint Admin role or Global Administrator role and be a site collection administrator
        to run the operation.

        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        :param str origin_url: Specifies a path to the doc library to be configured. It can be provided in two ways:
            relative path, or a mask.
        """
        payload = {
            "cdnType": cdn_type,
            "originUrl": origin_url,
        }
        qry = ServiceOperationQuery(self, "AddTenantCdnOrigin", None, payload)
        self.context.add_query(qry)
        return self

    def get_tenant_cdn_enabled(self, cdn_type):
        """
        Returns whether Public content delivery network (CDN) or Private CDN is enabled on the tenant level.

        You must have the SharePoint Admin role or Global Administrator role and be a site collection administrator
        to run the operation.

        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        """
        payload = {
            "cdnType": cdn_type,
        }
        return_type = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "GetTenantCdnEnabled", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def set_tenant_cdn_enabled(self, cdn_type, is_enabled):
        """
        Enables or disables Public content delivery network (CDN) or Private CDN on the tenant level.

        You must have the SharePoint Admin role or Global Administrator role and be a site collection administrator
        to run the operation.

        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        :param bool is_enabled: Specifies if the CDN is enabled.
        """
        payload = {
            "cdnType": cdn_type,
            "isEnabled": is_enabled
        }
        qry = ServiceOperationQuery(self, "SetTenantCdnEnabled", None, payload)
        self.context.add_query(qry)
        return self

    def remove_tenant_cdn_origin(self, cdn_type, origin_url):
        """
        Removes a new origin from the Public or Private content delivery network (CDN).

        You must have the SharePoint Admin role or Global Administrator role and be a site collection administrator
        to run the operation.

        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        :param str origin_url: Specifies a path to the doc library to be configured. It can be provided in two ways:
            relative path, or a mask.
        """
        payload = {
            "cdnType": cdn_type,
            "originUrl": origin_url,
        }
        qry = ServiceOperationQuery(self, "RemoveTenantCdnOrigin", None, payload)
        self.context.add_query(qry)
        return self

    def get_tenant_cdn_policies(self, cdn_type):
        """
        Get the public or private Policies applied on your SharePoint Online Tenant.

        Requires Tenant administrator permissions.


        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        """
        payload = {
            "cdnType": cdn_type,
        }
        return_type = ClientResult(self.context, ClientValueCollection(str))
        qry = ServiceOperationQuery(self, "GetTenantCdnPolicies", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def set_tenant_cdn_policy(self, cdn_type, policy, policy_value):
        """
        Sets the content delivery network (CDN) policies at the tenant level.

        Requires Tenant administrator permissions.

        :param int cdn_type: Specifies the CDN type. The valid values are: public or private.
        :param int policy: The PolicyType specifies the type of policy to set.
               Valid values:
                  IncludeFileExtensions
                  ExcludeRestrictedSiteClassifications
                  ExcludeIfNoScriptDisabled
        :param str policy_value: A String representing the value of the policy type defined by the PolicyType parameter.
        """
        payload = {
            "cdnType": cdn_type,
            "policy": policy,
            "policyValue": policy_value
        }
        qry = ServiceOperationQuery(self, "SetTenantCdnPolicy", None, payload)
        self.context.add_query(qry)
        return self

    def revoke_all_user_sessions(self, user_or_username):
        """
        Provides IT administrators the ability to invalidate a particular users' O365 sessions across all their devices.

        :param str or User user_or_username: Specifies a user name
              (for example, user1@contoso.com) or User object
        """
        return_type = SPOUserSessionRevocationResult(self.context)
        if isinstance(user_or_username, User):
            def _user_loaded():
                next_qry = ServiceOperationQuery(self, "RevokeAllUserSessions", [user_or_username.login_name], None,
                                                 None, return_type)
                self.context.add_query(next_qry)

            user_or_username.ensure_property("LoginName", _user_loaded)
        else:
            qry = ServiceOperationQuery(self, "RevokeAllUserSessions", [user_or_username], None, None, return_type)
            self.context.add_query(qry)
        return return_type

    def get_external_users(self, position=0, page_size=50, _filter=None, sort_order=0):
        """
        Returns external users in the tenant.

        :param int position: Use to specify the zero-based index of the position in the sorted collection of the
                             first result to be returned.
        :param int page_size: Specifies the maximum number of users to be returned in the collection.
                              The value must be less than or equal to 50.
        :param str _filter: Limits the results to only those users whose first name, last name, or email address
                            begins with the text in the string using a case-insensitive comparison.
        :param int sort_order: Specifies the sort results in Ascending or Descending order on the User.Email property
                               should occur.
        """
        return_type = GetExternalUsersResults(self.context)
        payload = {
            "position": position,
            "pageSize": page_size,
            "filter": _filter,
            "sortOrder": sort_order
        }
        qry = ServiceOperationQuery(self, "GetExternalUsers", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def remove_external_users(self, unique_ids=None):
        """
        Removes a collection of external users from the tenancy's folder.

        :param list[str] unique_ids: Specifies an ID that can be used to identify an external user based on
                                     their Windows Live ID.
        """
        payload = {
            "uniqueIds": unique_ids,
        }
        return_type = RemoveExternalUsersResults(self.context)
        qry = ServiceOperationQuery(self, "RemoveExternalUsers", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def get_all_tenant_themes(self):
        """
        Get all themes from tenant
        """
        return_type = ClientObjectCollection(self.context, ThemeProperties)
        qry = ServiceOperationQuery(self, "GetAllTenantThemes", None, None, None, return_type)
        self.context.add_query(qry)
        return return_type

    def add_tenant_theme(self, name, theme_json):
        """
        Adds a new theme to a tenant.

        :param str name:
        :param str theme_json:
        """
        return_type = ClientResult(self.context)
        payload = {
            "name": name,
            "themeJson": theme_json,
        }
        qry = ServiceOperationQuery(self, "AddTenantTheme", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def delete_tenant_theme(self, name):
        """
        Removes a theme from tenant

        :type name: str
        """
        payload = {
            "name": name,
        }
        qry = ServiceOperationQuery(self, "DeleteTenantTheme", None, payload)
        self.context.add_query(qry)
        return self



from office365.sharepoint.fields.field import Field


class FieldLookup(Field):
    """Specifies a lookup field."""

    @property
    def allow_multiple_values(self):
        """Gets a value that specifies whether the lookup field allows multiple values.
        :rtype: bool or None
        """
        return self.properties.get('AllowMultipleValues', None)

    @allow_multiple_values.setter
    def allow_multiple_values(self, value):
        """Sets a value that specifies whether the lookup field allows multiple values.
        """
        self.set_property('AllowMultipleValues', value)

    @property
    def lookup_web_id(self):
        """Gets the ID of the Web site that contains the list that is the source of this field's value.
        :rtype: str or None
        """
        return self.properties.get('LookupWebId', None)

    @lookup_web_id.setter
    def lookup_web_id(self, val):
        """Sets the ID of the Web site that contains the list that is the source of this field's value.
        :rtype: str or None
        """
        self.set_property("LookupWebId", val, True)

    @property
    def lookup_list(self):
        """Gets value that specifies the list identifier of the list that contains the field to use as the lookup
        values."""
        return self.properties.get('LookupList', None)

    @lookup_list.setter
    def lookup_list(self, val):
        """Sets a value that specifies the list identifier of the list that contains the field to use as
        the lookup values."""
        self.set_property("LookupList", val, True)



from office365.runtime.client_value import ClientValue
from office365.runtime.client_value_collection import ClientValueCollection
from office365.search.search_hit import SearchHit


class SearchHitsContainer(ClientValue):

    def __init__(self, hits=None, more_results_available=None, total=None):
        super(SearchHitsContainer, self).__init__()
        if hits is None:
            hits = ClientValueCollection(SearchHit)
        self.hits = hits
        self.moreResultsAvailable = more_results_available
        self.total = total



from office365.runtime.client_value import ClientValue


class ContentTypeCreationInformation(ClientValue):

    def __init__(self, name, description=None, group=None, ct_id=None):
        """Specifies properties that are used as parameters to initialize a new content type.
        :param str ct_id: Specifies the ContentTypeId (section 3.2.5.30) of the content type to be constructed.
        :param str name: Specifies the name of the content type to be constructed.
        :param str description: Specifies the description of the content type to be constructed.
        :param str group: Specifies the group of the content type to be constructed.

        """
        super(ContentTypeCreationInformation, self).__init__()
        self.Name = name
        self.Description = description
        self.group = group
        self.Id = ct_id



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity


class Reputation(BaseEntity):

    @staticmethod
    def set_rating(context, list_id, item_id, rating, return_type=None):
        """
        :param office365.sharepoint.client_context.ClientContext context: SharePoint context (client)
        :param str list_id: List Identifier
        :param int item_id: ListItem Identifier
        :param int rating: Rating number
        :param ClientResult return_type: return value
        """
        if return_type is None:
            return_type = ClientResult(context)

        binding_type = Reputation(context)
        payload = {
            "listID": list_id,
            "itemID": item_id,
            "rating": rating
        }
        qry = ServiceOperationQuery(binding_type, "SetRating", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @staticmethod
    def set_like(context, list_id, item_id, like, return_type=None):
        """
        :param office365.sharepoint.client_context.ClientContext context:
        :param str list_id: List Identifier
        :param int item_id: ListItem Identifier
        :param bool like: Like/Dislike value
        :param ClientResult return_type: return value
        """
        if return_type is None:
            return_type = ClientResult(context)
        binding_type = Reputation(context)
        payload = {
            "listID": list_id,
            "itemID": item_id,
            "like": like
        }
        qry = ServiceOperationQuery(binding_type, "SetLike", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @property
    def entity_type_name(self):
        return "Microsoft.Office.Server.ReputationModel.Reputation"




from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.outlook.outlook_category import OutlookCategory
from office365.runtime.resource_path import ResourcePath


class OutlookUser(Entity):
    """Represents the Outlook services available to a user."""

    def master_categories(self):
        return self.properties.get('masterCategories',
                                   EntityCollection(self.context, OutlookCategory,
                                                    ResourcePath("masterCategories", self.resource_path)))




from office365.onedrive.lists.list import List
from office365.runtime.client_object import ClientObject
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.webs.web import Web


class RemoteWeb(ClientObject):
    """Specifies a remote web that might be on a different domain."""

    def get_list_by_server_relative_url(self, serverRelativeUrl):
        target_list = List(self.context)
        qry = ServiceOperationQuery(self, "GetListByServerRelativeUrl", [serverRelativeUrl], None, None, target_list)
        self.context.add_query(qry)
        return target_list

    @staticmethod
    def create(context, requestUrl):
        remote_web = RemoteWeb(context)
        qry = ServiceOperationQuery(context, None, [requestUrl], None, None, remote_web)
        qry.static = True
        context.add_query(qry)
        return remote_web

    @property
    def web(self):
        """Gets the SPWeb."""
        return self.properties.get('Web', Web(self.context, ResourcePath("Web", self.resource_path)))



from office365.runtime.client_value import ClientValue


class TimeZoneInformation(ClientValue):
    """Provides information used to define a time zone."""

    def __init__(self, Bias=None, StandardBias=None, DaylightBias=None):
        """

        :param int Bias: Gets the bias in the number of minutes that the time zone differs from
            Coordinated Universal Time (UTC).
        :param DaylightBias: Gets the bias in the number of minutes that daylight time for the time zone
            differs from Coordinated Universal Time (UTC).
        :param StandardBias: Gets the bias in the number of minutes that standard time for the time zone differs
             from coordinated universal time (UTC).
        """
        super(TimeZoneInformation, self).__init__()
        self.Bias = Bias
        self.DaylightBias = DaylightBias
        self.StandardBias = StandardBias



from office365.entity import Entity


class IdentityProviderBase(Entity):
    """
    Represents identity providers with External Identities for both Azure Active Directory tenant and
    an Azure AD B2C tenant.
    """

    @property
    def display_name(self):
        """
        The display name for the identity provider.

        :rtype: str or None
        """
        return self.properties.get('displayName', None)



from office365.runtime.client_value import ClientValue


class SharingLink(ClientValue):
    """The SharingLink resource groups link-related data items into a single structure."""

    def __init__(self, _type=None, scope=None):
        """
        :param str _type: The type of the link created.
        :param str scope: The scope of the link represented by this permission. Value anonymous indicates the link is
             usable by anyone, organization indicates the link is only usable for users signed into the same tenant.
        """
        super(SharingLink, self).__init__()
        self.type = _type
        self.scope = scope



from office365.runtime.client_value import ClientValue


class AttachmentfileCreationInformation(ClientValue):

    def __init__(self, filename=None, content=None):
        """
        Represents properties that can be set when creating a file by using the AttachmentFiles.Add method.

        :type filename: str
        :type content: str or bytes
        """
        super(AttachmentfileCreationInformation, self).__init__()
        self._filename = filename
        self._content = content

    @property
    def content(self):
        """Gets the binary content of the file."""
        return self._content

    @content.setter
    def content(self, value):
        """Sets the binary content of the file."""
        self._content = value

    @property
    def filename(self):
        """The URL of the file."""
        return self._filename

    @filename.setter
    def filename(self, value):
        self._filename = value



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.publishing.topic_site_page import TopicSitePage


class TopicPageCollection(BaseEntityCollection):

    def __init__(self, context, resource_path=None):
        super(TopicPageCollection, self).__init__(context, TopicSitePage, resource_path)



from office365.runtime.client_value import ClientValue


class ContentTypeId(ClientValue):

    def __init__(self, string_value=None):
        """
        Represents the content type identifier (ID) of a content type.

        :param str string_value: Hexadecimal string value of content type identifier. String value MUST start with "0x".
        """
        super(ContentTypeId, self).__init__()
        self.StringValue = string_value



from office365.directory.users.password_profile import PasswordProfile
from office365.runtime.client_value import ClientValue


class UserProfile(ClientValue):
    def __init__(self, principal_name, password, display_name=None, account_enabled=False):
        """
        User profile

        :type principal_name: str
        :type password: str
        :type display_name: str
        :type account_enabled: bool
        """
        super(UserProfile, self).__init__()
        self.userPrincipalName = principal_name
        self.passwordProfile = PasswordProfile(password)
        self.mailNickname = principal_name.split("@")[0]
        self.displayName = display_name or principal_name.split("@")[0]
        self.accountEnabled = account_enabled



from office365.runtime.client_value import ClientValue


class Photo(ClientValue):
    """The photo resource provides photo and camera properties, for example, EXIF metadata, on a driveItem."""
    pass



from office365.onenote.operations.onenote_operation_error import OnenoteOperationError
from office365.onenote.operations.operation import Operation


class OnenoteOperation(Operation):
    """The status of certain long-running OneNote operations."""

    @property
    def error(self):
        """The error returned by the operation."""
        return self.properties.get("error", OnenoteOperationError())



from office365.runtime.client_object_collection import ClientObjectCollection
from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.resource_path import ResourcePath


class EntityCollection(ClientObjectCollection):
    """Microsoft Graph entity set"""

    def __getitem__(self, key):
        """

        :param key: key is used to address a Entity resource by either an index in collection
        or by resource id
        :type key: int or str
        :rtype: EntityCollection
        """
        if type(key) == int:
            return super(EntityCollection, self).__getitem__(key)
        return self._item_type(self.context, ResourcePath(key, self.resource_path))

    def new(self, **kwargs):
        return self.create_typed_object(properties=kwargs, persist_changes=True)

    def add(self, **kwargs):
        """Creates a resource

        :rtype: office365.entity.Entity
        """
        return_type = self.new()
        self.add_child(return_type)
        qry = CreateEntityQuery(self, kwargs, return_type)
        self.context.add_query(qry)
        return return_type

    @property
    def context(self):
        """
        :rtype: office365.graph_client.GraphClient
        """
        return self._context



from office365.entity_collection import EntityCollection
from office365.onenote.entity_hierarchy_model import OnenoteEntityHierarchyModel
from office365.onenote.operations.onenote_operation import OnenoteOperation
from office365.onenote.pages.page_links import PageLinks
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class OnenoteSection(OnenoteEntityHierarchyModel):
    """A section in a OneNote notebook. Sections can contain pages."""

    def copy_to_section_group(self, group_id, _id, rename_as=None, site_collection_id=None, site_id=None):
        """For Copy operations, you follow an asynchronous calling pattern: First call the Copy action,
        and then poll the operation endpoint for the result.

        :param str group_id: The id of the group to copy to. Use only when copying to a Microsoft 365 group.
        :param str _id: Required. The id of the destination section group.
        :param str rename_as: The name of the copy. Defaults to the name of the existing item.
        :param str site_collection_id:
        :param str site_id:
        """
        return_type = OnenoteOperation(self.context)
        payload = {
            "groupId": group_id,
            "id": _id,
            "renameAs": rename_as,
            "siteCollectionId": site_collection_id,
            "siteId": site_id
        }
        qry = ServiceOperationQuery(self, "copyToSectionGroup", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    @property
    def is_default(self):
        """Indicates whether this is the user's default section. Read-only."""
        return self.properties.get("isDefault", None)

    @property
    def links(self):
        """Links for opening the section. The oneNoteClientURL link opens the section in the OneNote native client
        if it's installed. The oneNoteWebURL link opens the section in OneNote on the web.

        """
        return self.properties.get("links", PageLinks())

    @property
    def pages(self):
        """
        The collection of pages in the section. Read-only. Nullable.

        :rtype: EntityCollection
        """
        from office365.onenote.pages.page import OnenotePage
        return self.get_property('pages',
                                 EntityCollection(self.context, OnenotePage, ResourcePath("pages", self.resource_path)))

    @property
    def parent_notebook(self):
        """The notebook that contains the page. Read-only.

        :rtype: Notebook
        """
        from office365.onenote.notebooks.notebook import Notebook
        return self.get_property('parentNotebook',
                                 Notebook(self.context, ResourcePath("parentNotebook", self.resource_path)))

    @property
    def parent_section_group(self):
        """The section group that contains the section. Read-only.

        :rtype: SectionGroup
        """
        from office365.onenote.sectiongroups.section_group import SectionGroup
        return self.get_property('parentSectionGroup',
                                 SectionGroup(self.context, ResourcePath("parentSectionGroup", self.resource_path)))

    def get_property(self, name, default_value=None):
        if default_value is None:
            property_mapping = {
                "parentNotebook": self.parent_notebook,
                "parentSectionGroup": self.parent_section_group
            }
            default_value = property_mapping.get(name, None)
        return super(OnenoteSection, self).get_property(name, default_value)



from office365.runtime.client_value import ClientValue


class MoveCopyOptions(ClientValue):

    def __init__(self, keep_both=True,
                 reset_author_and_created_on_copy=False,
                 retain_editor_and_modified_on_move=False,
                 should_bypass_shared_locks=False):
        super(MoveCopyOptions, self).__init__()
        self.KeepBoth = keep_both
        self.ResetAuthorAndCreatedOnCopy = reset_author_and_created_on_copy
        self.RetainEditorAndModifiedOnMove = retain_editor_and_modified_on_move
        self.ShouldBypassSharedLocks = should_bypass_shared_locks

    @property
    def entity_type_name(self):
        return "SP.MoveCopyOptions"



import os
import uuid
from xml.etree import ElementTree
import xml.dom.minidom as minidom

import requests
import requests.utils

import office365.logger
from office365.runtime.auth.authentication_provider import AuthenticationProvider
from office365.runtime.auth.sts_profile import STSProfile
from office365.runtime.auth.user_realm_info import UserRealmInfo

office365.logger.ensure_debug_secrets()


def resolve_base_url(url):
    parts = url.split('://')
    host_name = parts[1].split("/")[0]
    return parts[0] + '://' + host_name


def xml_escape(s_val):
    s_val = s_val.replace("&", "&amp;")
    s_val = s_val.replace("<", "&lt;")
    s_val = s_val.replace(">", "&gt;")
    s_val = s_val.replace("\"", "&quot;")
    s_val = s_val.replace("'", "&apos;")
    return s_val


def is_valid_auth_cookies(values):
    return any(values) and (values.get('FedAuth', None) is not None or values.get('SPOIDCRL', None) is not None)


class SamlTokenProvider(AuthenticationProvider, office365.logger.LoggerContext):

    def __init__(self, url, username, password, browser_mode):
        """SAML Security Token Service provider

        :type url: str
        :type username: str
        :type password: str
        :type browser_mode: str
        """
        # Security Token Service info
        self._sts_profile = STSProfile(resolve_base_url(url))
        # Obtain authentication cookies, using the browser mode
        self._browser_mode = browser_mode
        # Last occurred error
        self.error = ''
        self._username = username
        self._password = password
        self._cached_auth_cookies = None
        self.__ns_prefixes = {
            'S': '{http://www.w3.org/2003/05/soap-envelope}',
            's': '{http://www.w3.org/2003/05/soap-envelope}',
            'psf': '{http://schemas.microsoft.com/Passport/SoapServices/SOAPFault}',
            'wst': '{http://schemas.xmlsoap.org/ws/2005/02/trust}',
            'wsse': '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd}',
            'saml': '{urn:oasis:names:tc:SAML:1.0:assertion}',
            'u': '{http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd}',
            'wsa': '{http://www.w3.org/2005/08/addressing}',
            'wsp': '{http://schemas.xmlsoap.org/ws/2004/09/policy}',
            'ps': '{http://schemas.microsoft.com/LiveID/SoapServices/v1}',
            'ds': '{http://www.w3.org/2000/09/xmldsig#}'
        }
        for key in self.__ns_prefixes.keys():
            ElementTree.register_namespace(key, self.__ns_prefixes[key][1:-1])

    def authenticate_request(self, request):
        """

        :param request:
        """
        logger = self.logger(self.authenticate_request.__name__)
        self.ensure_authentication_cookie()
        logger.debug_secrets(self._cached_auth_cookies)
        cookie_header_value = "; ".join(["=".join([key, str(val)]) for key, val in self._cached_auth_cookies.items()])
        request.set_header('Cookie', cookie_header_value)

    def ensure_authentication_cookie(self):
        if self._cached_auth_cookies is None:
            self._cached_auth_cookies = self.get_authentication_cookie()
        return True

    def get_authentication_cookie(self):
        """Acquire authentication cookie
        """
        logger = self.logger(self.ensure_authentication_cookie.__name__)
        logger.debug('get_authentication_cookie called')

        try:
            logger.debug("Acquiring Access Token..")
            user_realm = self._get_user_realm()
            if user_realm.IsFederated:
                token = self._acquire_service_token_from_adfs(user_realm.STSAuthUrl)
            else:
                token = self._acquire_service_token()
            return self._get_authentication_cookie(token, user_realm.IsFederated)
        except requests.exceptions.RequestException as e:
            logger.error(e.response.text)
            self.error = "Error: {}".format(e)
            raise ValueError(e.response.text)

    def _get_user_realm(self):
        """Get User Realm
        """
        resp = requests.post(self._sts_profile.user_realm_service_url, data="login={0}&xml=1".format(self._username),
                             headers={'Content-Type': 'application/x-www-form-urlencoded'})
        xml = ElementTree.fromstring(resp.content)
        node = xml.find('NameSpaceType')
        if node is not None:
            if node.text == 'Federated':
                info = UserRealmInfo(xml.find('STSAuthURL').text, True)
            else:
                info = UserRealmInfo(None, False)
            return info
        return None

    def get_last_error(self):
        return self.error

    def _acquire_service_token_from_adfs(self, adfs_url):
        logger = self.logger(self._acquire_service_token_from_adfs.__name__)

        payload = self._prepare_request_from_template('FederatedSAML.xml', {
            'auth_url': adfs_url,
            'message_id': str(uuid.uuid4()),
            'username': self._username,
            'password': self._password,
            'created': self._sts_profile.created,
            'expires': self._sts_profile.expires,
            'issuer': self._sts_profile.tokenIssuer
        })

        response = requests.post(adfs_url, data=payload,
                                 headers={'Content-Type': 'application/soap+xml; charset=utf-8'})
        dom = minidom.parseString(response.content.decode())
        assertion_node = dom.getElementsByTagNameNS("urn:oasis:names:tc:SAML:1.0:assertion", 'Assertion')[0].toxml()

        try:
            payload = self._prepare_request_from_template('RST2.xml', {
                'auth_url': self._sts_profile.tenant,
                'serviceTokenUrl': self._sts_profile.security_token_service_url,
                'assertion_node': assertion_node
            })

            # 3. get security token
            response = requests.post(self._sts_profile.security_token_service_url, data=payload,
                                     headers={'Content-Type': 'application/soap+xml'})
            token = self._process_service_token_response(response)
            logger.debug_secrets('security token: %s', token)
            return token
        except ElementTree.ParseError as e:
            self.error = 'An error occurred while parsing the server response: {}'.format(e)
            logger.error(self.error)
            return None

    def _acquire_service_token(self):
        """Retrieve service token"""
        logger = self.logger(self._acquire_service_token.__name__)

        payload = self._prepare_request_from_template('SAML.xml', {
            'auth_url': self._sts_profile.authorityUrl,
            'username': xml_escape(self._username),
            'password': xml_escape(self._password),
            'message_id': str(uuid.uuid4()),
            'created': self._sts_profile.created,
            'expires': self._sts_profile.expires,
            'issuer': self._sts_profile.tokenIssuer
        })
        logger.debug_secrets('options: %s', payload)
        response = requests.post(self._sts_profile.security_token_service_url, data=payload,
                                 headers={'Content-Type': 'application/x-www-form-urlencoded'})
        token = self._process_service_token_response(response)
        logger.debug_secrets('security token: %s', token)
        return token

    def _process_service_token_response(self, response):
        logger = self.logger(self._process_service_token_response.__name__)
        logger.debug_secrets('response: %s\nresponse.content: %s', response, response.content)

        try:
            xml = ElementTree.fromstring(response.content)
        except ElementTree.ParseError as e:
            self.error = 'An error occurred while parsing the server response: {}'.format(e)
            logger.error(self.error)
            return None

        # check for errors
        if xml.find('{0}Body/{0}Fault'.format(self.__ns_prefixes['s'])) is not None:
            error = xml.find(
                '{0}Body/{0}Fault/{0}Detail/{1}error/{1}internalerror/{1}text'.format(self.__ns_prefixes['s'],
                                                                                      self.__ns_prefixes['psf']))
            if error is None:
                self.error = 'An error occurred while retrieving token from XML response.'
            else:
                self.error = 'An error occurred while retrieving token from XML response: {0}'.format(error.text)
            logger.error(self.error)
            raise ValueError(self.error)

        # extract token
        token = xml.find(
            '{0}Body/{1}RequestSecurityTokenResponse/{1}RequestedSecurityToken/{2}BinarySecurityToken'.format(
                self.__ns_prefixes['s'], self.__ns_prefixes['wst'], self.__ns_prefixes['wsse']))
        if token is None:
            self.error = 'Cannot get binary security token for from {0}'.format(
                self._sts_profile.security_token_service_url)
            logger.error(self.error)
            raise ValueError(self.error)
        logger.debug_secrets("token: %s", token)
        return token.text

    def _get_authentication_cookie(self, security_token, federated=False):
        """Retrieve auth cookie from STS

        :type federated: bool
        :type security_token: str
        """
        logger = self.logger(self._get_authentication_cookie.__name__)

        session = requests.session()
        logger.debug_secrets("session: %s\nsession.post(%s, data=%s)", session, self._sts_profile.signin_page_url,
                             security_token)
        if not federated or self._browser_mode:
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            if self._browser_mode:
                headers['User-Agent'] = 'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)'
            session.post(self._sts_profile.signin_page_url, data=security_token, headers=headers)
        else:
            idcrl_endpoint = "https://{}/_vti_bin/idcrl.svc/".format(self._sts_profile.tenant)
            session.get(idcrl_endpoint,
                        headers={
                            'User-Agent': 'Office365 Python Client',
                            'X-IDCRL_ACCEPTED': 't',
                            'Authorization': 'BPOSIDCRL {0}'.format(security_token)
                        })
        logger.debug_secrets("session.cookies: %s", session.cookies)
        cookies = requests.utils.dict_from_cookiejar(session.cookies)
        logger.debug_secrets("cookies: %s", cookies)
        if not is_valid_auth_cookies(cookies):
            self.error = "An error occurred while retrieving auth cookies from {0}".format(
                self._sts_profile.signin_page_url)
            logger.error(self.error)
            raise ValueError(self.error)
        return cookies

    @staticmethod
    def _prepare_request_from_template(template_name, params):
        """Construct the request body to acquire security token from STS endpoint"""
        logger = SamlTokenProvider.logger()
        logger.debug_secrets('params: %s', params)
        f = open(os.path.join(os.path.dirname(__file__), "templates", template_name))
        try:
            data = f.read()
            for key in params:
                data = data.replace('{' + key + '}', str(params[key]))
            return data
        finally:
            f.close()



from office365.runtime.client_result import ClientResult
from office365.runtime.client_value_collection import ClientValueCollection
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath
from office365.sharepoint.base_entity import BaseEntity
from office365.sharepoint.search.query.popularTenantQuery import PopularTenantQuery
from office365.sharepoint.search.query.querySuggestionResults import QuerySuggestionResults
from office365.sharepoint.search.searchResult import SearchResult


class SearchService(BaseEntity):

    def __init__(self, context):
        super(SearchService, self).__init__(context, ResourcePath("Microsoft.Office.Server.Search.REST.SearchService"))

    def export(self, user_name, start_time):
        """
        The operation is used by the administrator to retrieve the query log entries,
        issued after a specified date, for a specified user.

        :param datetime.datetime start_time: The timestamp of the oldest query log entry returned.
        :param str user_name: The name of the user that issued the queries."""
        result = ClientResult(self.context)
        payload = {
            "userName": user_name,
            "startTime": start_time.isoformat()
        }
        qry = ServiceOperationQuery(self, "export", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def export_manual_suggestions(self):
        """"""
        pass

    def export_popular_tenant_queries(self, count):
        result = ClientResult(self.context, ClientValueCollection(PopularTenantQuery))
        payload = {
            "count": count,
        }
        qry = ServiceOperationQuery(self, "exportpopulartenantqueries", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def query(self, search_request):
        """The operation is used to retrieve search results by using the HTTP protocol with the GET method.

        :type search_request: office365.sharepoint.search.searchRequest.SearchRequest
        """
        result = ClientResult(self.context, SearchResult())
        qry = ServiceOperationQuery(self, "query", search_request.to_json(), None, "query", result)
        self.context.add_query(qry)

        def _construct_request(request):
            """
            :type request: office365.runtime.http.request_options.RequestOptions
            """
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_request)
        return result

    def post_query(self, request):
        """The operation is used to retrieve search results through the use of the HTTP protocol
        with method type POST.

        :type request: office365.sharepoint.search.searchRequest.SearchRequest"""
        result = ClientResult(self.context, SearchResult())
        qry = ServiceOperationQuery(self, "postquery", None, request, "request", result)
        self.context.add_query(qry)
        return result

    def record_page_click(self):
        """The operation is used to record page clicks"""
        pass

    def search_center_url(self):
        """The operation is used to get the URI address of the search center by using the HTTP protocol
        with the GET method. The operation returns the URI of the of the search center."""
        result = ClientResult(self.context)
        qry = ServiceOperationQuery(self, "searchCenterUrl", None, None, None, result)
        self.context.add_query(qry)
        return result

    def results_page_address(self):
        """The operation is used to get the URI address of the result page by using the HTTP protocol
        with the GET method. The operation returns the URI of the result page."""
        pass

    def suggest(self, query_text):
        result = ClientResult(self.context, QuerySuggestionResults())
        payload = {
            "querytext": query_text
        }
        qry = ServiceOperationQuery(self, "suggest", None, payload, None, result)
        self.context.add_query(qry)
        return result

    def auto_completions(self):
        pass



from office365.runtime.client_value import ClientValue
from office365.sharepoint.views.view_scope import ViewScope


class CamlQuery(ClientValue):

    def __init__(self, datesInUtc=True, viewXml=None, listItemCollectionPosition=None,
                 folderServerRelativeUrl=None, allowIncrementalResults=True):
        """
        Specifies a Collaborative Application Markup Language (CAML) query on a list or joined lists.

        :type bool allowIncrementalResults: Specifies whether the incremental results can be returned.
        :param ListItemCollectionPosition listItemCollectionPosition: Specifies the information required to
            get the next page of data for the list view.
        :param str viewXml: Specifies the XML schema that defines the list view.
        :param str or None folderServerRelativeUrl: Specifies the server-relative URL of a list folder from which
            results are to be returned.
        :param bool datesInUtc: Specifies whether the query returns dates in Coordinated Universal Time (UTC) format.
        """
        super(CamlQuery, self).__init__()
        self.DatesInUtc = datesInUtc
        self.FolderServerRelativeUrl = folderServerRelativeUrl
        self.AllowIncrementalResults = allowIncrementalResults
        self.ViewXml = viewXml
        self.ListItemCollectionPosition = listItemCollectionPosition

    @staticmethod
    def parse(query_expr, scope=ViewScope.DefaultValue):
        """
        Construct CamlQuery object from expression
        :type query_expr: str
        :type scope: ViewScope
        """
        qry = CamlQuery()
        qry.ViewXml = "<View Scope=\"{0}\"><Query>{1}</Query></View>".format(scope, query_expr)
        return qry

    @staticmethod
    def create_custom_query(query):
        qry = CamlQuery()
        qry.ViewXml = query
        return qry

    @staticmethod
    def create_all_items_query():
        return CamlQuery.parse("", ViewScope.RecursiveAll)

    @staticmethod
    def create_all_folders_query():
        qry_text = "<Where><Eq><FieldRef Name=\"FSObjType\" /><Value Type=\"Integer\">1</Value></Eq></Where>"
        return CamlQuery.parse(qry_text, ViewScope.RecursiveAll)

    @staticmethod
    def create_all_files_query():
        qry_text = "<Where><Eq><FieldRef Name=\"FSObjType\" /><Value Type=\"Integer\">0</Value></Eq></Where>"
        return CamlQuery.parse(qry_text, ViewScope.RecursiveAll)

    @property
    def entity_type_name(self):
        return 'SP.CamlQuery'



from office365.runtime.client_value import ClientValue


class UserIdentity(ClientValue):
    """
    In the context of an Azure AD audit log, this represents the user information that initiated or
    was affected by an audit activity.
    """

    def __init__(self):
        super(UserIdentity, self).__init__()
        self.displayName = None
        self.ipAddress = None
        self.userPrincipalName = None



class ListTemplateType:
    def __init__(self):
        pass

    InvalidType = -1
    NoListTemplate = 0
    GenericList = 100
    DocumentLibrary = 101
    Survey = 102
    Links = 103
    Announcements = 104
    Contacts = 105
    Events = 106
    Tasks = 107
    DiscussionBoard = 108
    PictureLibrary = 109
    DataSources = 110
    WebTemplateCatalog = 111
    UserInformation = 112
    WebPartCatalog = 113
    ListTemplateCatalog = 114
    XMLForm = 115
    MasterPageCatalog = 116
    NoCodeWorkflows = 117
    WorkflowProcess = 118
    WebPageLibrary = 119
    CustomGrid = 120
    SolutionCatalog = 121
    NoCodePublic = 122
    ThemeCatalog = 123
    DesignCatalog = 124
    AppDataCatalog = 125
    DataConnectionLibrary = 130
    WorkflowHistory = 140
    GanttTasks = 150
    HelpLibrary = 151
    AccessRequest = 160
    TasksWithTimelineAndHierarchy = 171
    MaintenanceLogs = 175
    Meetings = 200
    Agenda = 201
    MeetingUser = 202
    Decision = 204
    MeetingObjective = 207
    TextBox = 210
    ThingsToBring = 211
    HomePageLibrary = 212
    Posts = 301
    Comments = 302
    Categories = 303
    Facility = 402
    Whereabouts = 403
    CallTrack = 404
    Circulation = 405
    Timecard = 420
    Holidays = 421
    IMEDic = 499



from office365.runtime.odata.odata_path_parser import ODataPathParser
from office365.runtime.resource_path import ResourcePath


class ResourcePathServiceOperation(ResourcePath):
    """ Resource path to address Service Operations which
    represents simple functions exposed by an OData service"""

    def __init__(self, method_name, method_parameters=None, parent=None):
        """
        :type method_parameters: list or dict or office365.runtime.client_value.ClientValue or None
        :type method_name: str


        """
        super(ResourcePathServiceOperation, self).__init__(ODataPathParser.from_method(method_name, method_parameters),
                                                           parent)

    @property
    def delimiter(self):
        return ""



from office365.runtime.client_value_collection import ClientValueCollection
from office365.sharepoint.changes.change import Change
from office365.sharepoint.sharing.sharedWithUser import SharedWithUser


class ChangeItem(Change):
    """A change on an item."""

    @property
    def shared_with_users(self):
        return self.properties.get("SharedWithUsers", ClientValueCollection(SharedWithUser))



from office365.runtime.client_value import ClientValue


class AssignedLicense(ClientValue):
    """"""

    def __init__(self, sku_id=None, disabled_plans=None):
        super(AssignedLicense, self).__init__()
        self.skuId = sku_id
        self.disabledPlans = disabled_plans



from office365.runtime.queries.create_entity_query import CreateEntityQuery
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.principal.user import User


class UserCollection(BaseEntityCollection):
    """Represents a collection of User resources."""

    def __init__(self, context, resource_path=None):
        super(UserCollection, self).__init__(context, User, resource_path)

    def get(self):
        """
        :rtype: UserCollection
        """
        return super(UserCollection, self).get()

    def add_user(self, login_name):
        """
        Adds a user

        :type login_name: str
        """
        user = User(self.context)
        user._parent_collection = self
        user.set_property('LoginName', login_name)
        qry = CreateEntityQuery(self, user, user)
        self.context.add_query(qry)
        self.add_child(user)
        return user

    def get_by_email(self, email):
        """Retrieve User object by email

        :type email: str
        """
        return User(self.context, ResourcePathServiceOperation("GetByEmail", [email], self.resource_path))

    def get_by_id(self, user_id):
        """Retrieve User object by id"""
        return User(self.context, ResourcePathServiceOperation("GetById", [user_id], self.resource_path))

    def get_by_login_name(self, login_name):
        """Retrieve User object by login name

        :type login_name: str
        """
        return User(self.context,
                    ResourcePathServiceOperation("GetByLoginName", [login_name], self.resource_path))

    def remove_by_id(self, _id):
        """Retrieve User object by id"""
        qry = ServiceOperationQuery(self, "RemoveById", [_id])
        self.context.add_query(qry)

    def remove_by_login_name(self, login_name):
        """Remove User object by login name"""
        qry = ServiceOperationQuery(self, "RemoveByLoginName", [login_name])
        self.context.add_query(qry)



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity


class WebApplication(BaseEntity):

    @staticmethod
    def lookup(context, request_uri):
        """
        :type context
        :type request_uri str
        """
        return_type = WebApplication(context)
        payload = {"requestUri": request_uri}
        qry = ServiceOperationQuery(return_type, "Lookup", None, payload, None, return_type)
        qry.static = True
        context.add_query(qry)
        return return_type

    @property
    def entity_type_name(self):
        return "Microsoft.SharePoint.Administration.SPWebApplication"



from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.listitems.listitem import ListItem


class ListItemCollection(BaseEntityCollection):
    """List Item collection"""
    def __init__(self, context, resource_path=None):
        super(ListItemCollection, self).__init__(context, ListItem, resource_path)



from office365.entity import Entity


class ConditionalAccessPolicy(Entity):
    """
    Represents an Azure Active Directory conditional access policy.
    Conditional access policies are custom rules that define an access scenario.
    """
    pass



from office365.runtime.client_value import ClientValue


class File(ClientValue):
    """
    The File resource groups file-related data items into a single structure.

    If a DriveItem has a non-null file facet, the item represents a file.
    In addition to other properties, files have a content relationship which contains the byte stream of the file.
    """

    def __init__(self, mime_type=None, hashes=None):
        """
        :param str mime_type: The MIME type for the file. This is determined by logic on the server and might not be
            the value provided when the file was uploaded. Read-only.

        :param str hashes: Hashes of the file's binary content, if available. Read-only.
        """
        super(File, self).__init__()
        self.hashes = hashes
        self.mimeType = mime_type
        self.processingMetadata = None



from office365.base_item import BaseItem
from office365.directory.identities.identity_set import IdentitySet
from office365.entity_collection import EntityCollection
from office365.onedrive.driveitems.driveItem import DriveItem
from office365.onedrive.lists.list import List
from office365.onedrive.internal.resourcepaths.root_resource_path import RootResourcePath
from office365.onedrive.driveitems.system_facet import SystemFacet
from office365.runtime.http.http_method import HttpMethod
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path import ResourcePath


class Drive(BaseItem):
    """The drive resource is the top level object representing a user's OneDrive or a document library in
    SharePoint. """

    def search(self, query_text):
        """Search the hierarchy of items for items matching a query.

        :type query_text: str
        """
        return_type = EntityCollection(self.context, DriveItem, ResourcePath("items", self.resource_path))
        qry = ServiceOperationQuery(self, "search", {"q": query_text}, None, None, return_type)
        self.context.add_query(qry)

        def _construct_request(request):
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_request)
        return return_type

    def recent(self):
        """
        List a set of items that have been recently used by the signed in user.
        This collection includes items that are in the user's drive as well as items
        they have access to from other drives.
        """
        return_type = EntityCollection(self.context, DriveItem, ResourcePath("items", self.resource_path))
        qry = ServiceOperationQuery(self, "recent", None, None, None, return_type)
        self.context.add_query(qry)

        def _construct_request(request):
            request.method = HttpMethod.Get

        self.context.before_execute(_construct_request)
        return return_type

    @property
    def drive_type(self):
        """
        Describes the type of drive represented by this resource. OneDrive personal drives will return personal.
        OneDrive for Business will return business. SharePoint document libraries will return documentLibrary.

        :rtype: str or None
        """
        return self.properties.get("driveType", None)

    @property
    def sharepoint_ids(self):
        """Returns identifiers useful for SharePoint REST compatibility."""
        return self.properties.get('sharepointIds', None)

    @property
    def system(self):
        """Optional. The user account that owns the drive. Read-only."""
        return self.properties.get('system', SystemFacet())

    @property
    def owner(self):
        """If present, indicates that this is a system-managed drive. Read-only."""
        return self.properties.get('owner', IdentitySet())

    @property
    def shared_with_me(self):
        """Retrieve a collection of DriveItem resources that have been shared with the owner of the Drive.

        :rtype: EntityCollection
        """
        return self.get_property('sharedWithMe',
                                 EntityCollection(self.context, DriveItem,
                                                  ResourcePath("sharedWithMe", self.resource_path)))

    @property
    def root(self):
        """The root folder of the drive.

        :rtype: DriveItem
        """
        return self.get_property('root', DriveItem(self.context, RootResourcePath(self.resource_path)))

    @property
    def list(self):
        """For drives in SharePoint, the underlying document library list.

        :rtype: List
        """
        return self.get_property('list', List(self.context, ResourcePath("list", self.resource_path)))

    @property
    def items(self):
        """All items contained in the drive.

        :rtype: EntityCollection
        """
        return self.get_property('items',
                                 EntityCollection(self.context, DriveItem, ResourcePath("items", self.resource_path)))

    @property
    def following(self):
        """The list of items the user is following. Only in OneDrive for Business.

        :rtype: EntityCollection
        """
        return self.get_property('following',
                                 EntityCollection(self.context, DriveItem,
                                                  ResourcePath("following", self.resource_path)))

    @property
    def special(self):
        """Collection of common folders available in OneDrive. Read-only. Nullable.

        :rtype: EntityCollection
        """
        return self.get_property('special',
                                 EntityCollection(self.context, DriveItem,
                                                  ResourcePath("special", self.resource_path)))



from office365.directory.directory_object import DirectoryObject
from office365.runtime.resource_path import ResourcePath


class DirectoryRole(DirectoryObject):
    """Represents an Azure AD directory role. Azure AD directory roles are also known as administrator roles """

    @property
    def members(self):
        """Users that are members of this directory role. HTTP Methods: GET, POST, DELETE. Read-only. Nullable.

        :rtype: DirectoryObjectCollection
        """
        from office365.directory.directory_object_collection import DirectoryObjectCollection
        return self.get_property('members',
                                 DirectoryObjectCollection(self.context, ResourcePath("members", self.resource_path)))



from office365.entity import Entity


class PlannerTask(Entity):
    """
    The plannerTask resource represents a Planner task in Microsoft 365.
    A Planner task is contained in a plan and can be assigned to a bucket in a plan.
    Each task object has a details object which can contain more information about the task.
    See overview for more information regarding relationships between group, plan and task.
    """
    pass



from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.runtime.resource_path_service_operation import ResourcePathServiceOperation
from office365.sharepoint.base_entity_collection import BaseEntityCollection
from office365.sharepoint.files.file_version import FileVersion


class FileVersionCollection(BaseEntityCollection):
    """Represents a collection of FileVersion."""
    def __init__(self, context, resource_path=None):
        super(FileVersionCollection, self).__init__(context, FileVersion, resource_path)

    def get_by_id(self, version_id):
        """Gets the file version with the specified ID."""
        return FileVersion(self.context, ResourcePathServiceOperation("getById", [version_id], self.resource_path))

    def delete_all(self):
        """Deletes all the file version objects in the collection."""
        qry = ServiceOperationQuery(self, "DeleteAll")
        self.context.add_query(qry)
        return self

    def delete_by_id(self, vid):
        """Removes the file version object with the specified integer ID from the collection.

        :param int vid: The file version to remove.
        """
        qry = ServiceOperationQuery(self, "DeleteByID", {"vid": vid})
        self.context.add_query(qry)
        return self

    def delete_by_label(self, label):
        """
        Deletes the file version object with the specified version label.

        :param str label: The file version to remove.
        """
        qry = ServiceOperationQuery(self, "DeleteByLabel", {"versionlabel": label})
        self.context.add_query(qry)
        return self

    def recycle_by_id(self, vid):
        """
        Recycles a file version objects in the collection by version identifier.

        :param int vid: The file version to remove.
        """
        qry = ServiceOperationQuery(self, "RecycleByID", {"vid": vid})
        self.context.add_query(qry)
        return self

    def recycle_by_label(self, label):
        """
        Recycles the file version object with the specified version label.

        :param str label: The file version to remove.
        """
        qry = ServiceOperationQuery(self, "RecycleByLabel", {"versionlabel": label})
        self.context.add_query(qry)
        return self

    def restore_by_label(self, label):
        """
        Restores the file version object that has the specified version label.

        :param str label: The file version to remove.
        """
        qry = ServiceOperationQuery(self, "RestoreByLabel", {"versionlabel": label})
        self.context.add_query(qry)
        return self



class PrincipalSource:
    """Specifies the source of a principal."""

    def __init__(self):
        pass

    None_ = 0
    """Do not specify a principal source."""

    UserInfoList = 1
    """Use the user information list as the source."""

    Windows = 2
    """Use Windows as the source."""

    MembershipProvider = 4
    """Use the current membership provider as the source."""

    RoleProvider = 8
    """Use the current role provider as the source"""

    All = 15
    """Use all principal sources."""



from office365.planner.tasks.check_list_item import PlannerChecklistItem
from office365.runtime.client_value_collection import ClientValueCollection


class PlannerChecklistItems(ClientValueCollection):
    """The plannerChecklistItemCollection resource represents the collection of checklist items on a task.
    It is an Open Type. It is part of the task details object.
    The value in the property-value pair is the checklistItem object.
    """

    def __init__(self, initial_values=None):
        super(PlannerChecklistItems, self).__init__(PlannerChecklistItem, initial_values)



from office365.runtime.client_result import ClientResult
from office365.runtime.queries.service_operation_query import ServiceOperationQuery
from office365.sharepoint.base_entity import BaseEntity


class ThemeManager(BaseEntity):
    """SharePoint site theming"""

    @property
    def entity_type_name(self):
        return "SP.Utilities.ThemeManager"

    def add_tenant_theme(self, name, theme_json):
        """
        Adds a new theme to a tenant.

        :param str name:
        :param str theme_json:
        """
        return_type = ClientResult(self.context)
        payload = {
            "name": name,
            "themeJson": theme_json,
        }
        qry = ServiceOperationQuery(self, "AddTenantTheme", None, payload, None, return_type)
        self.context.add_query(qry)
        return return_type

    def delete_tenant_theme(self, name):
        """
        Removes a theme.
        """
        payload = {
            "name": name,
        }
        qry = ServiceOperationQuery(self, "DeleteTenantTheme", None, payload)
        self.context.add_query(qry)
        return self



from office365.entity import Entity


class PlannerPlan(Entity):
    """The plannerPlan resource represents a plan in Microsoft 365. A plan can be owned by a group
    and contains a collection of plannerTasks. It can also have a collection of plannerBuckets.
    Each plan object has a details object that can contain more information about the plan.
    For more information about the relationships between groups, plans, and tasks, see Planner.
    """
    pass



from office365.entity import Entity
from office365.entity_collection import EntityCollection
from office365.planner.plans.plan import PlannerPlan
from office365.runtime.resource_path import ResourcePath


class Planner(Entity):
    """
    The planner resource is the entry point for the Planner object model.
    It returns a singleton planner resource. It doesn't contain any usable properties.
    """

    @property
    def plans(self):
        """Read-only. Nullable. Returns the plannerTasks assigned to the user.

        :rtype: EntityCollection
        """
        return self.get_property('plans',
                                 EntityCollection(self.context, PlannerPlan,
                                                  ResourcePath("plans", self.resource_path)))



def _normalize(key, value):
    if key == "select" or key == "expand":
        return ",".join(value)
    return value


class QueryOptions(object):

    def __init__(self, select=None, expand=None, filter_expr=None, orderBy=None, top=None, skip=None):
        """
        A query option is a set of query string parameters applied to a resource that can help control the amount
        of data being returned for the resource in the URL

        :param list[str] select: The $select system query option allows the clients to requests a limited set of
        properties for each entity or complex type.
        :param list[str] expand: The $expand system query option specifies the related resources to be included in
        line with retrieved resources.
        :param str filter_expr: The $filter system query option allows clients to filter a collection of resources
        that are addressed by a request URL.
        :param str orderBy: The $orderby system query option allows clients to request resources in either ascending
        order using asc or descending order using desc
        :param int top: The $top system query option requests the number of items in the queried collection to
        be included in the result.
        :param int skip: The $skip query option requests the number of items in the queried collection that
        are to be skipped and not included in the result.
        """
        if expand is None:
            expand = []
        if select is None:
            select = []
        self.select = select
        self.expand = expand
        self.filter = filter_expr
        self.orderBy = orderBy
        self.skip = skip
        self.top = top

    @property
    def is_empty(self):
        result = {k: v for (k, v) in self.__dict__.items() if v is not None and v}
        return not result

    def reset(self):
        self.select = []
        self.expand = []
        self.filter = None
        self.orderBy = None
        self.skip = None
        self.top = None

    def to_url(self):
        """Convert query options to url
        :return: str
        """
        return '&'.join(['$%s=%s' % (key, _normalize(key, value))
                         for (key, value) in self.__dict__.items() if value is not None and value])
